# Copyright (c) 2006-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
This example shows how to build a torus cluster with multi-core hosts.

However, each leaf in the torus is a StarZone, composed of several CPUs

Each actor runs in a specific CPU. One sender broadcasts a message to all receivers.
"""

import sys
import typing
import simgrid


class Sender:
    """
    Send a msg for each host in its host list
    """

    def __init__(self, hosts, msg_size=int(1e6)):
        self.hosts = hosts
        self.msg_size = msg_size

    # Actors that are created as object will execute their __call__ method.
    # So, the following constitutes the main function of the Sender actor.
    def __call__(self):
        pending_comms = []
        mboxes = []

        for host in self.hosts:
            msg = "Hello, I'm alive and running on " + simgrid.this_actor.get_host().name
            mbox = simgrid.Mailbox.by_name(host.name)
            mboxes.append(mbox)
            pending_comms.append(mbox.put_async(msg, self.msg_size))

        simgrid.this_actor.info("Done dispatching all messages")

        # Now that all message exchanges were initiated, wait for their completion in one single call
        simgrid.Comm.wait_all(pending_comms)

        simgrid.this_actor.info("Goodbye now!")


class Receiver:
    """
    Receiver actor: wait for 1 message on the mailbox identified by the hostname
    """

    def __call__(self):
        mbox = simgrid.Mailbox.by_name(simgrid.this_actor.get_host().name)
        received = mbox.get()
        simgrid.this_actor.info("I got a '%s'." % received)

#####################################################################################################


def create_hostzone(zone: simgrid.NetZone, coord: typing.List[int], ident: int) -> typing.Tuple[simgrid.NetPoint,
                                                                                                simgrid.NetPoint]:
    r"""
    Callback to set a cluster leaf/element

    In our example, each leaf if a StarZone, composed of 8 CPUs.
    Each CPU is modeled as a host, connected to the outer world through a high-speed PCI link.
    Obs.: CPU0 is the gateway for this zone

       (outer world)
            CPU0 (gateway)
       up ->|   |
            |   |<-down
            +star+
         /   / \   \
        /   /   \   \<-- 100Gbs, 10us link (1 link UP and 1 link DOWN for full-duplex)
       /   /     \   \
      /   /       \   \
      CPU1   ...   CPU8

    :param zone: Cluster netzone being created (usefull to create the hosts/links inside it)
    :param coord: Coordinates in the cluster
    :param ident: Internal identifier in the torus (for information)
    :return netpoint, gateway: the netpoint to the StarZone and CPU0 as gateway
    """
    num_cpus = 8     # Number of CPUs in the zone
    speed = "1Gf"    # Speed of each CPU
    link_bw = "100GBps"  # Link bw connecting the CPU
    link_lat = "1ns"  # Link latency

    hostname = "host" + str(ident)
    # create the StarZone
    host_zone = simgrid.NetZone.create_star_zone(hostname)
    # setting my Torus parent zone
    host_zone.set_parent(zone)

    gateway = None
    # create CPUs
    for i in range(num_cpus):
        cpu_name = hostname + "-cpu" + str(i)
        host = host_zone.create_host(cpu_name, speed).seal()
        # the first CPU is the gateway
        if i == 0:
            gateway = host
        # create split-duplex link
        link = host_zone.create_split_duplex_link("link-" + cpu_name, link_bw)
        link.set_latency(link_lat).seal()
        # connecting CPU to outer world
        host_zone.add_route(host.netpoint, None, None, None,
                            [simgrid.LinkInRoute(link, simgrid.LinkInRoute.Direction.UP)], True)

    # seal newly created netzone
    host_zone.seal()
    return host_zone.netpoint, gateway.netpoint

#####################################################################################################


def create_limiter(zone: simgrid.NetZone, coord: typing.List[int], ident: int) -> simgrid.Link:
    """
    Callback to create limiter link (1Gbs) for each netpoint

    The coord parameter depends on the cluster being created:
    - Torus: Direct translation of the Torus' dimensions, e.g. (0, 0, 0) for a 3-D Torus
    - Fat-Tree: A pair (level in the tree, ident), e.g. (0, 0) for first leaf in the tree and (1,0) for the first switch
    at level 1.
    - Dragonfly: a tuple (group, chassis, blades/routers, nodes), e.g. (0, 0, 0, 0) for first node in the cluster. To
    identify the router inside a (group, chassis, blade), we use MAX_UINT in the last parameter (e.g. 0, 0, 0,
    4294967295).

    :param zone: Torus netzone being created (usefull to create the hosts/links inside it)
    :param coord: Coordinates in the cluster
    :param ident: Internal identifier in the torus (for information)
    :return: Limiter link
    """
    return zone.create_link("limiter-" + str(ident), [1e9]).seal()


def create_torus_cluster():
    """
    Creates a TORUS cluster

    Creates a TORUS cluster with dimensions 2x2x2

    The cluster has 8 elements/leaves in total. Each element is a StarZone containing 8 Hosts.
    Each pair in the torus is connected through 2 links:
    1) limiter: a 1Gbs limiter link (set by user through the set_limiter callback)
    2) link: 10Gbs link connecting the components (created automatically)

    (Y-axis=2)
    A
    |
    |   D (Z-axis=2)
    +  / 10 Gbs
    | +
    |/ limiter=1Gps
    B-----+----C (X-axis=2)

    For example, a communication from A to C goes through:
    <tt> A->limiter(A)->link(A-B)->limiter(B)->link(B-C)->limiter(C)->C </tt>

    More precisely, considering that A and C are StarZones, a
    communication from A-CPU-3 to C-CPU-7 goes through:
    1) StarZone A: A-CPU-3 -> link-up-A-CPU-3 -> A-CPU-0
    2) A-CPU-0->limiter(A)->link(A-B)->limiter(B)->link(B-C)->limiter(C)->C-CPU-0
    3) StarZone C: C-CPU-0-> link-down-C-CPU-7 -> C-CPU-7

    Note that we don't have limiter links inside the StarZones(A, B, C),
    but we have limiters in the Torus that are added to the links in the path (as we can see in "2)")

    More details in: <a href="https://simgrid.org/doc/latest/Platform_examples.html?highlight=torus#torus-cluster">Torus
    Cluster</a>
    """
    # create the torus cluster, 10Gbs link between elements in the cluster
    simgrid.NetZone.create_torus_zone("cluster", None, [2, 2, 2],
                                      simgrid.ClusterCallbacks(create_hostzone, None, create_limiter), 10e9, 10e-6,
                                      simgrid.Link.SharingPolicy.SPLITDUPLEX).seal()

#####################################################################################################


def create_fat_tree_cluster():
    r"""
    Creates a Fat-Tree cluster

    Creates a Fat-Tree cluster with 2 levels and 6 nodes
    The following parameters are used to create this cluster:
    - Levels: 2 - two-level of switches in the cluster
    - Down links: 2, 3 - L2 routers is connected to 2 elements, L1 routers to 3 elements
    - Up links: 1, 2 - Each node (A-F) is connected to 1 L1 router, L1 routers are connected to 2 L2
    - Link count: 1, 1 - Use 1 link in each level

    The first parameter describes how many levels we have.
    The following ones describe the connection between the elements and must have exactly n_levels components.


                            S3     S4                <-- Level 2 routers
       link:limiter -      /   \  /  \
                          +     ++    +
       link: 10GBps -->  |     /  \    |
        (full-duplex)    |    /    \   |
                         +   +      +  +
                         |  /        \ |
                         S1           S2             <-- Level 1 routers
      link:limiter ->    |             |
                         +             +
     link:10GBps  -->   /|\           /|\
                       / | \         / | \
                      +  +  +       +  +  +
     link:limiter -> /   |   \     /   |   \
                    A    B    C   D    E    F        <-- level 0 Nodes

    Each element (A to F) is a StarZone containing 8 Hosts.
    The connection uses 2 links:
    1) limiter: a 1Gbs limiter link (set by user through the set_limiter callback)
    2) link: 10Gbs link connecting the components (created automatically)

    For example, a communication from A to C goes through:
    <tt> A->limiter(A)->link(A-S1)->limiter(S1)->link(S1-C)->->limiter(C)->C</tt>

    More precisely, considering that A and C are StarZones, a
    communication from A-CPU-3 to C-CPU-7 goes through:
    1) StarZone A: A-CPU-3 -> link-up-A-CPU-3 -> A-CPU-0
    2) A-CPU-0->limiter(A)->link(A-S1)->limiter(S1)->link(S1-C)->limiter(C)->C-CPU-0
    3) StarZone C: C-CPU-0-> link-down-C-CPU-7 -> C-CPU-7

    More details in: <a href="https://simgrid.org/doc/latest/Platform_examples.html#fat-tree-cluster">Fat-Tree
    Cluster</a>
    """
    # create the fat tree cluster, 10Gbs link between elements in the cluster
    simgrid.NetZone.create_fatTree_zone("cluster", None, simgrid.FatTreeParams(2, [2, 3], [1, 2], [1, 1]),
                                        simgrid.ClusterCallbacks(create_hostzone, None, create_limiter), 10e9, 10e-6,
                                        simgrid.Link.SharingPolicy.SPLITDUPLEX).seal()

#####################################################################################################


def create_dragonfly_cluster():
    r"""
    Creates a Dragonfly cluster

    Creates a Dragonfly cluster with 2 groups and 16 nodes
    The following parameters are used to create this cluster:
    - Groups: 2 groups, connected with 2 links (blue links)
    - Chassis: 2 chassis, connected with a single link (black links)
    - Routers: 2 routers, connected with 2 links (green links)
    - Nodes: 2 leaves per router, single link

    The diagram below illustrates a group in the dragonfly cluster

    +------------------------------------------------+
    |        black link(1)                           |
    |     +------------------------+                 |
    | +---|--------------+     +---|--------------+  |
    | |   |  green       |     |   |  green       |  |
    | |   |  links (2)   |     |   |  links (2)   |  |   blue links(2)
    | |   R1 ====== R2   |     |   R3 -----  R4 ======================> "Group 2"
    | |  /  \      /  \  |     |  /  \      /  \  |  |
    | | A    B    C    D |     | E    F    G    H |  |
    | +------------------+     +------------------+  |
    |      Chassis 1                Chassis 2        |
    +------------------------------------------------+
     Group 1

    Each element (A, B, C, etc) is a StarZone containing 8 Hosts.
    The connection between elements (e.g. A->R1) uses 2 links:
    1) limiter: a 1Gbs limiter link (set by user through the set_limiter callback)
    2) link: 10Gbs link connecting the components (created automatically)

    For example, a communication from A to C goes through:
    <tt> A->limiter(A)->link(A-R1)->limiter(R1)->link(R1-R2)->limiter(R2)->link(R2-C)limiter(C)->C</tt>

    More details in: <a href="https://simgrid.org/doc/latest/Platform_examples.html#dragonfly-cluster">Dragonfly
    Cluster</a>
    """
    # create the dragonfly cluster, 10Gbs link between elements in the cluster
    simgrid.NetZone.create_dragonfly_zone("cluster", None, simgrid.DragonflyParams([2, 2], [2, 1], [2, 2], 2),
                                          simgrid.ClusterCallbacks(create_hostzone, None, create_limiter), 10e9, 10e-6,
                                          simgrid.Link.SharingPolicy.SPLITDUPLEX).seal()

###################################################################################################


def main():
    e = simgrid.Engine(sys.argv)
    platform = sys.argv[1]

    # create platform
    if platform == "torus":
        create_torus_cluster()
    elif platform == "fatTree":
        create_fat_tree_cluster()
    elif platform == "dragonfly":
        create_dragonfly_cluster()
    else:
        sys.exit("invalid param")

    host_list = e.all_hosts
    # create the sender actor running on first host
    simgrid.Actor.create("sender", host_list[0], Sender(host_list))
    # create receiver in every host
    for host in host_list:
        simgrid.Actor.create("receiver-" + host.name, host, Receiver())

    # runs the simulation
    e.run()

if __name__ == '__main__':
    main()



# Copyright (c) 2007-2022. The SimGrid Team. All rights reserved.          
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package. 

"""
 This is a version of the masterworkers that (hopefully) survives to the chaos monkey.
 It tests synchronous send/receive as well as synchronous computations.

 It is not written to be pleasant to read, but instead to resist the aggressions of the monkey:
 - Workers keep going until after a global variable `todo` reaches 0.
 - The master is a daemon that just sends infinitely tasks
   (simgrid simulations stop as soon as all non-daemon actors are done).
 - The platform is created programmatically to remove path issues and control the problem size.

 See the simgrid-monkey script for more information.

 Inline configuration items:
 - host-count: how many actors to start (including the master
 - task-count: initial value of the `todo` global
 - deadline: time at which the simulation is known to be failed (to detect infinite loops).

"""

# Configuration items:
host_count = 3 # Host count (master on one, workers on the others)
task_count = 1 # Amount of tasks that must be executed to succeed
deadline = 120 # When to fail the simulation (infinite loop detection)
# End of configuration

import sys
from simgrid import Actor, Engine, Host, this_actor, Mailbox, NetZone, LinkInRoute, TimeoutException, NetworkFailureException

todo = task_count # remaining amount of tasks to execute, a global variable

def master():
  comp_size = int(1e6)
  comm_size = int(1e6)
  this_actor.info("Master booting")
  Actor.self().daemonize()
  this_actor.on_exit(lambda killed: this_actor.info("Master dying forcefully." if killed else "Master dying peacefully."))

  while True: # This is a daemon
    assert Engine.clock < deadline, f"Failed to run all tasks in less than {deadline} seconds. Is this an infinite loop?"

    try: 
      this_actor.info("Try to send a message")
      mailbox.put(comp_size, comm_size, 10.)
    except TimeoutException:
      this_actor.info("Timeouted while sending a task")
    except NetworkFailureException:
      this_actor.info("Got a NetworkFailureException. Wait a second before starting again.")
      this_actor.sleep_for(1.)

  assert False, "The impossible just happened (yet again): daemons shall not finish."

def worker(my_id):
  global todo
  this_actor.info(f"Worker {my_id} booting")
  this_actor.on_exit(lambda killed: this_actor.info(f"Worker {my_id} dying {'forcefully' if killed else 'peacefully'}."))

  while todo > 0:
    assert Engine.clock < deadline, f"Failed to run all tasks in less than {deadline} seconds. Is this an infinite loop?"

    try:
      this_actor.info(f"Waiting a message on mailbox")
      compute_cost = mailbox.get()

      this_actor.info("Start execution...")
      this_actor.execute(compute_cost)
      todo = todo - 1
      this_actor.info(f"Execution complete. Still {todo} to go.")

    except NetworkFailureException:
      this_actor.info("Got a NetworkFailureException. Wait a second before starting again.")
      this_actor.sleep_for(1.)
    except TimeoutException:
      this_actor.info("Timeouted while getting a task.")

if __name__ == '__main__':
  e = Engine(sys.argv)

  assert host_count > 2, "You need at least 2 workers (i.e., 3 hosts) or the master will be auto-killed when the only worker gets killed."
  assert todo > 0, "Please give some tasks to do to the workers."

  mailbox = Mailbox.by_name("mailbox")

  rootzone = NetZone.create_full_zone("Zone1")
  main = rootzone.create_host("lilibeth 0", 1e9)
  Actor.create("master", main, master).set_auto_restart(True)

  for i in range(1, host_count):
    link = rootzone.create_split_duplex_link(f"link {i}", "1MBps").set_latency("24us")
    host = rootzone.create_host(f"lilibeth {i}", 1e9)
    rootzone.add_route(main.netpoint, host.netpoint, None, None, [LinkInRoute(link, LinkInRoute.Direction.UP)], True)
    Actor.create("worker", host, worker, i).set_auto_restart(True)

  e.netzone_root.seal()
  e.run()

  this_actor.info("WE SURVIVED!")



# Copyright (c) 2017-2022. The SimGrid Team. All rights reserved.
#
# This program is free software you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
Usage: actor-suspend.py platform_file [other parameters]
"""

import sys
from simgrid import Actor, Engine, this_actor


def lazy_guy():
    """The Lazy guy only wants to sleep, but can be awaken by the dream_master actor"""
    this_actor.info("Nobody's watching me ? Let's go to sleep.")
    this_actor.suspend()  # - Start by suspending itself
    this_actor.info("Uuuh ? Did somebody call me ?")

    # - Then repetitively go to sleep, but get awaken
    this_actor.info("Going to sleep...")
    this_actor.sleep_for(10)
    this_actor.info("Mmm... waking up.")

    this_actor.info("Going to sleep one more time (for 10 sec)...")
    this_actor.sleep_for(10)
    this_actor.info("Waking up once for all!")

    this_actor.info("Ok, let's do some work, then (for 10 sec on Boivin).")
    this_actor.execute(980.95e6)

    this_actor.info("Mmmh, I'm done now. Goodbye.")


def dream_master():
    """The Dream master"""
    this_actor.info("Let's create a lazy guy.")  # Create a lazy_guy actor
    lazy = Actor.create("Lazy", this_actor.get_host(), lazy_guy)
    this_actor.info("Let's wait a little bit...")
    this_actor.sleep_for(10)  # Wait for 10 seconds
    this_actor.info("Let's wake the lazy guy up! >:) BOOOOOUUUHHH!!!!")
    if lazy.is_suspended():
        lazy.resume()  # Then wake up the lazy_guy
    else:
        this_actor.error(
            "I was thinking that the lazy guy would be suspended now")

    this_actor.sleep_for(5)  # Repeat two times:
    this_actor.info("Suspend the lazy guy while he's sleeping...")
    lazy.suspend()  # Suspend the lazy_guy while he's asleep
    this_actor.info("Let him finish his siesta.")
    this_actor.sleep_for(10)  # Wait for 10 seconds
    this_actor.info("Wake up, lazy guy!")
    lazy.resume()  # Then wake up the lazy_guy again

    this_actor.sleep_for(5)
    this_actor.info("Suspend again the lazy guy while he's sleeping...")
    lazy.suspend()
    this_actor.info("This time, don't let him finish his siesta.")
    this_actor.sleep_for(2)
    this_actor.info("Wake up, lazy guy!")
    lazy.resume()

    this_actor.sleep_for(5)
    this_actor.info(
        "Give a 2 seconds break to the lazy guy while he's working...")
    lazy.suspend()
    this_actor.sleep_for(2)
    this_actor.info("Back to work, lazy guy!")
    lazy.resume()

    this_actor.info("OK, I'm done here.")


if __name__ == '__main__':
    e = Engine(sys.argv)
    if len(sys.argv) < 2:
        raise AssertionError(
            "Usage: actor-suspend.py platform_file [other parameters]")

    e.load_platform(sys.argv[1])  # Load the platform description
    hosts = e.all_hosts
    Actor.create("dream_master", hosts[0], dream_master)

    e.run()  # Run the simulation



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: MBI

  Description: @{shortdesc}@
    @{longdesc}@

  Version of MPI: Conforms to MPI 1.1, does not require MPI 2 implementation

BEGIN_MPI_FEATURES
  P2P!basic: @{p2pfeature}@
  P2P!nonblocking: @{ip2pfeature}@
  P2P!persistent: Lacking
  COLL!basic: Lacking
  COLL!nonblocking: Lacking
  COLL!persistent: Lacking
  COLL!tools: Lacking
  RMA: Lacking
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>


int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;
  MPI_Status sta;
  int src,dest;
  int stag=0, rtag=0;
  int buff_size = 1;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug!\\n");

  MPI_Comm newcom = MPI_COMM_WORLD;
  MPI_Datatype type = MPI_INT;

  @{init1a}@
  @{init1b}@
  @{init1c}@
  @{init2a}@
  @{init2b}@
  @{init2c}@

  if (rank == 0) {
    dest=1, src=1;
    @{operation1a}@ /* MBIERROR1 */
    @{operation1b}@
    @{operation1c}@
    @{fini1a}@
    @{fini1b}@
    @{fini1c}@
  }else if (rank == 1){
    dest=0, src=0;
    @{operation2a}@ /* MBIERROR2 */
    @{operation2b}@
    @{operation2c}@
    @{fini2a}@
    @{fini2b}@
    @{fini2c}@
  }
  @{free1a}@
  @{free1b}@
  @{free1c}@
  @{free2a}@
  @{free2b}@
  @{free2c}@

  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""


for p in gen.probe:
    for s in gen.send + gen.isend:
        for r in gen.recv + gen.irecv:
            patterns = {}
            patterns = {'p':p, 's': s, 'r': r}
            patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
            patterns['p2pfeature'] = 'Yes' if s in gen.send or r in gen.recv else 'Lacking'
            patterns['ip2pfeature'] = 'Yes' if s in gen.isend or r in gen.irecv else 'Lacking'
            patterns['s'] = s
            patterns['r'] = r
            patterns['p'] = p
            patterns['init1a'] = gen.init[p]("1")
            patterns['init1b'] = gen.init[s]("1")
            patterns['init1c'] = gen.init[r]("2")
            patterns['init2a'] = gen.init[p]("1")
            patterns['init2b'] = gen.init[r]("3")
            patterns['init2c'] = gen.init[s]("4")
            patterns['fini1a'] = gen.fini[p]("1")
            patterns['fini1b'] = gen.fini[s]("1")
            patterns['fini1c'] = gen.fini[r]("2")
            patterns['fini2a'] = gen.fini[p]("1")
            patterns['fini2b'] = gen.fini[r]("3")
            patterns['fini2c'] = gen.fini[s]("4")
            patterns['free1a'] = gen.free[p]("1")
            patterns['free1b'] = gen.free[s]("1")
            patterns['free1c'] = gen.free[r]("2")
            patterns['free2a'] = gen.free[p]("1")
            patterns['free2b'] = gen.free[r]("3")
            patterns['free2c'] = gen.free[s]("4")
            patterns['operation1a'] = gen.operation[p]("1")
            patterns['operation1b'] = gen.operation[s]("1")
            patterns['operation1c'] = gen.operation[r]("2")
            patterns['operation2a'] = gen.operation[p]("1")
            patterns['operation2b'] = gen.operation[r]("3")
            patterns['operation2c'] = gen.operation[s]("4")

            # Generate the incorrect matching
            replace = patterns.copy()
            replace['shortdesc'] = 'MPI_Probe is called before MPI_Recv.'
            replace['longdesc'] = 'MPI_Probe is a blocking call that returns only after a matching message has been found. By calling MPI_Probe before MPI_Recv, a deadlock is created.'
            replace['outcome'] = 'ERROR: CallMatching'
            replace['errormsg'] = 'P2P mistmatch. @{p}@ at @{filename}@:@{line:MBIERROR1}@ and @{filename}@:@{line:MBIERROR2}@ are called before @{r}@.'
            gen.make_file(template, f'CallOrdering_{p}_{r}_{s}_nok.c', replace)

            # Generate a correct matching
            replace = patterns.copy()
            replace['shortdesc'] = 'Correct use of MPI_Probe.'
            replace['longdesc'] = 'Correct use of MPI_Probe.'
            replace['outcome'] = 'OK'
            replace['errormsg'] = 'OK'
            replace['operation1a'] = gen.operation[s]("1")
            replace['operation1b'] = gen.operation[p]("1")
            gen.make_file(template, f'CallOrdering_{p}_{r}_{s}_ok.c', replace)



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: @{origin}@

  Description: @{shortdesc}@
    @{longdesc}@


BEGIN_MPI_FEATURES
  P2P!basic: Lacking
  P2P!nonblocking: Lacking
  P2P!persistent: Lacking
  COLL!basic: Lacking
  COLL!nonblocking: Lacking
  COLL!persistent: Lacking
  COLL!tools: Lacking
  RMA: @{rmafeature}@
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

#define N 10

int * buffer;

static void get_win(MPI_Win *win) {
  @{bufferalloc}@

  MPI_Win_create(@{buffer}@, N * sizeof(int), 1, MPI_INFO_NULL, MPI_COMM_WORLD, win);

  return;
}

int main(int argc, char *argv[]) {
  int rank, numProcs;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &numProcs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (numProcs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug!\\n");

  MPI_Win win;

  get_win(&win);

  MPI_Win_fence(0, win);

  if (rank == 0) {
    int localbuf[N] = {0};
    localbuf[0] = 12345;
    MPI_Put(&localbuf, N, MPI_INT, 1, 0, N, MPI_INT, win);
  }

  MPI_Win_fence(0, win);

  MPI_Win_free(&win);

  @{bufferfree}@

  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}

"""


for b in ['missing', 'null', 'malloc', 'bufferSize']:
    patterns = {}
    patterns = {'b': b}
    patterns['origin'] = "MPI-CorrBench"
    patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
    patterns['rmafeature'] = 'Yes'

    replace = patterns.copy()
    replace['shortdesc'] = 'Invalid buffer in window creation.'
    replace['longdesc'] = 'Invalid buffer in window creation.'
    replace['outcome'] = 'ERROR: InvalidBuffer'
    replace['errormsg'] = '@{b}@ at @{filename}@:@{line:MBIERROR}@ has an invalid buffer'
    replace['bufferfree'] = ''

    ok = 'nok'
    replace['buffer'] = 'buffer'

    if b == 'missing':
        replace['bufferalloc'] = '/* MBIERROR1 */'
        replace['longdesc'] = 'Uninitialized buffer in window creation.'
    elif b == 'null':
        replace['bufferalloc'] = 'buffer = NULL; /* MBIERROR1 */'
        replace['longdesc'] = 'Use NULL buffer in window creation.'
    elif b == 'bufferSize':
        replace['bufferalloc'] = 'buffer = malloc((N/2) * sizeof(int)); /* MBIERROR1 */'
        replace['bufferfree'] = 'free(buffer);'
        replace['longdesc'] = 'Unmatched size of buffer in window creation.'
    else:
        replace['bufferalloc'] = 'buffer = malloc(N * sizeof(int));'
        replace['bufferfree'] = 'free(buffer);'
        replace['longdesc'] = 'Correct initialized buffer in window creation.'
        replace['outcome'] = 'OK'
        replace['errormsg'] = ''
        ok = 'ok'

    gen.make_file(template, f'InvalidParam_WinBuffer_{b}_{ok}.c', replace)



#!/usr/bin/env python

# Copyright (c) 2013-2022. The SimGrid Team.
# All rights reserved.

# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
This script generates a specific deployment file for the Bittorrent example.
It assumes that the platform will be a cluster.
Usage: python generate.py nb_nodes nb_bits end_date percentage
Example: python generate.py 10000 5000
"""

import sys
import random

if len(sys.argv) != 4:
    print(
        "Usage: python generate.py nb_nodes end_date seed_percentage > deployment_file.xml")
    sys.exit(1)

nb_nodes = int(sys.argv[1])
end_date = int(sys.argv[2])
seed_percentage = int(sys.argv[3])

nb_bits = 24
max_id = 2 ** nb_bits - 1
all_ids = [42]

sys.stdout.write("<?xml version='1.0'?>\n"
                 "<!DOCTYPE platform SYSTEM \"https://simgrid.org/simgrid.dtd\">\n"
                 "<platform version=\"3\">\n"
                 "  <process host=\"c-0.me\" function=\"bittorrent.Tracker\"><argument value=\"%d\"/></process>\n" %
                 end_date)

for i in range(1, nb_nodes):

    ok = False
    while not ok:
        my_id = random.randint(0, max_id)
        ok = my_id not in all_ids
    start_date = i * 10
    line = "  <process host=\"c-%d.me\" function=\"bittorrent.Peer\"><argument value=\"%d\" />" \
           "<argument value=\"%d\" />" % (i, my_id, end_date)
    if random.randint(0, 100) < seed_percentage:
        line += "<argument value=\"1\" />"
    line += "</process>\n"
    sys.stdout.write(line)
    all_ids.append(my_id)
sys.stdout.write("</platform>")



# Copyright (c) 2018-2022. The SimGrid Team. All rights reserved.
#
# This program is free software you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
Usage: exec-async.py platform_file [other parameters]
"""

import sys
from simgrid import Actor, Engine, Host, this_actor


class Waiter:
    """
    This actor simply waits for its task completion after starting it.
    That's exactly equivalent to synchronous execution.
    """

    def __call__(self):
        computation_amount = this_actor.get_host().speed
        this_actor.info("Waiter executes {:.0f} flops, should take 1 second.".format(computation_amount))
        activity = this_actor.exec_init(computation_amount)
        activity.start()
        activity.wait()

        this_actor.info("Goodbye from waiter!")


class Monitor:
    """This actor tests the ongoing execution until its completion, and don't wait before it's terminated."""

    def __call__(self):
        computation_amount = this_actor.get_host().speed
        this_actor.info("Monitor executes {:.0f} flops, should take 1 second.".format(computation_amount))
        activity = this_actor.exec_init(computation_amount).start()

        while not activity.test():
            this_actor.info("Remaining amount of flops: {:.0f} ({:.0f}%)".format(
                activity.remaining, 100 * activity.remaining_ratio))
            this_actor.sleep_for(0.3)
        activity.wait()

        this_actor.info("Goodbye from monitor!")


class Canceller:
    """This actor cancels the ongoing execution after a while."""

    def __call__(self):
        computation_amount = this_actor.get_host().speed
        this_actor.info("Canceller executes {:.0f} flops, should take 1 second.".format(computation_amount))
        activity = this_actor.exec_async(computation_amount)

        this_actor.sleep_for(0.5)
        this_actor.info("I changed my mind, cancel!")
        activity.cancel()

        this_actor.info("Goodbye from canceller!")


if __name__ == '__main__':
    e = Engine(sys.argv)
    if len(sys.argv) < 2:
        raise AssertionError("Usage: exec-async.py platform_file [other parameters]")

    e.load_platform(sys.argv[1])

    Actor.create("wait", Host.by_name("Fafard"), Waiter())
    Actor.create("monitor", Host.by_name("Ginette"), Monitor())
    Actor.create("cancel", Host.by_name("Boivin"), Canceller())

    e.run()



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: MBI

  Description: @{shortdesc}@
    @{longdesc}@

  Version of MPI: Conforms to MPI 1.1, does not require MPI 2 implementation

BEGIN_MPI_FEATURES
  P2P!basic: @{p2pfeature}@
  P2P!nonblocking: @{ip2pfeature}@
  P2P!persistent: Lacking
  COLL!basic: Lacking
  COLL!nonblocking: Lacking
  COLL!persistent: Lacking
  COLL!tools: Lacking
  RMA: Lacking
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 4 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>


int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;
  int src=MPI_ANY_SOURCE, dest=0;
  int stag = 42, rtag = MPI_ANY_TAG;
  int buff_size = 1;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug!\\n");

  int send_buffer=rank;

  MPI_Datatype type = MPI_INT;
  MPI_Comm newcom = MPI_COMM_WORLD;

  @{init1}@
  @{init2}@

  if (rank == 0) {
    for (int i = 0; i < nprocs - 1; i++) {
      @{operation1}@ /* MBIERROR */
      @{fini1}@
    }
  if (@{cond}@ != 3) {
      printf("MBI_MSG_RACE: The last received message is not 3 but %d!\\n", buf1);
      fflush(stdout);
      abort();
    }
  }else{
    @{operation2}@
    @{fini2}@
  }


  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""


for s in gen.send + gen.isend:
    for r in gen.recv + gen.irecv:
        patterns = {}
        patterns = {'s': s, 'r': r}
        patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
        patterns['p2pfeature'] = 'Yes' if s in gen.send or r in gen.recv else 'Lacking'
        patterns['ip2pfeature'] = 'Yes' if s in gen.isend or r in gen.irecv else 'Lacking'
        patterns['s'] = s
        patterns['r'] = r
        patterns['cond'] = 'buf1'
        patterns['init2'] = gen.init[s]("2")
        patterns['init1'] = gen.init[r]("1")
        patterns['fini2'] = gen.fini[s]("2")
        patterns['fini1'] = gen.fini[r]("1")
        patterns['operation2'] = gen.operation[s]("2")
        patterns['operation1'] = gen.operation[r]("1")

        # Generate the incorrect matching
        replace = patterns.copy()
        replace['shortdesc'] = 'The message ordering is non-deterministic.'
        replace['longdesc'] = 'The code assumes a fixed order in the reception of messages while the message ordering is non-deterministic.'
        replace['outcome'] = 'ERROR: MessageRace'
        replace['errormsg'] = 'P2P message race which can cause a deadlock. @{r}@ at @{filename}@:@{line:MBIERROR}@ is called with ANY_SRC.'
        gen.make_file(template, f'MessageRace_{r}_{s}_nok.c', replace)



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
This example demonstrates how to attach a profile to a host or a link, to specify external changes to the resource
speed.
The first way to do so is to use a file in the XML, while the second is to use the programmatic interface.
"""

import sys
from simgrid import Actor, Engine, Host, Link, this_actor

def watcher():
    jupiter = Host.by_name("Jupiter")
    fafard = Host.by_name("Fafard")
    lilibeth = Host.by_name("Lilibeth")
    link1 = Link.by_name("1")
    link2 = Link.by_name("2")

    (links, lat) = jupiter.route_to(fafard)
    path = ""
    for l in links:
        path += ("" if not path else ", ") + "link '" + l.name + "'"
    this_actor.info(f"Path from Jupiter to Fafard: {path} (latency: {lat:.6f}s).")

    for _ in range(10):
        this_actor.info("Fafard: %.0fMflops, Jupiter: %4.0fMflops, Lilibeth: %3.1fMflops, \
Link1: (%.2fMB/s %.0fms), Link2: (%.2fMB/s %.0fms)" % (fafard.speed * fafard.available_speed / 1000000,
                                                       jupiter.speed * jupiter.available_speed / 1000000,
                                                       lilibeth.speed * lilibeth.available_speed / 1000000,
                                                       link1.bandwidth / 1000, link1.latency * 1000,
                                                       link2.bandwidth / 1000, link2.latency * 1000))
        this_actor.sleep_for(1)

if __name__ == '__main__':
    e = Engine(sys.argv)
    # Load the platform description
    e.load_platform(sys.argv[1])

    # Add a new host programmatically, and attach a simple speed profile to it (alternate between full and half speed
    # every two seconds
    lili = e.netzone_root.create_host("Lilibeth", 25e6)
    lili.set_speed_profile("""0 1.0
    2 0.5""", 4)
    lili.seal()

    # Add a watcher of the changes
    Actor.create("watcher", Host.by_name("Fafard"), watcher)

    e.run()



#
# Copyright 2012-2015 Bronto Software, Inc. and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Convert Java syntax tree nodes to string representations.

"""

import javalang

from .util import StringBuilder

# The order for displaying modifiers
__modifiers_order = ('public', 'protected', 'private', 'static', 'abstract', 'final',
                     'native', 'synchronized', 'transient', 'volatile', 'strictfp')

def formatter(f):
    def _f(node, output=None, **kwargs):
        if output is None:
            output = StringBuilder()

        f(node, output, **kwargs)
        return output
    return _f

def output_list(f, items, output=None, sep=', '):
    if items:
        f(items[0], output)
        for item in items[1:]:
            output.append(sep)
            f(item, output)

@formatter
def output_annotation(annotation, output):
    output.append('@')
    output.append(annotation.name)
    output.append(' ')

@formatter
def output_type(typ, output, with_generics=True):
    if not typ:
        output.append('void')
        return

    if typ.dimensions:
        dim = '[]' * len(typ.dimensions)
    else:
        dim = ''

    if isinstance(typ, javalang.tree.BasicType):
        output.append(typ.name)
    else:
        while typ:
            output.append(typ.name)

            if with_generics:
                output_type_args(typ.arguments, output)

            typ = typ.sub_type

            if typ:
                output.append('.')
    output.append(dim)

@formatter
def output_exception(exception, output):
    output.append(exception)

@formatter
def output_type_arg(type_arg, output):
    if type_arg.pattern_type == '?':
        output.append('?')
    else:
        if type_arg.pattern_type:
            output.append('? ')
            output.append(type_arg.pattern_type)
            output.append(' ')

        output_type(type_arg.type, output)

@formatter
def output_type_args(type_args, output):
    if type_args:
        output.append('<')
        output_list(output_type_arg, type_args, output, ', ')
        output.append('>')

@formatter
def output_type_param(type_param, output):
    output.append(type_param.name)

    if type_param.extends:
        output.append(' extends ')
        output_list(output_type, type_param.extends, output, ' & ')

@formatter
def output_type_params(type_params, output):
    if type_params:
        output.append('<')
        output_list(output_type_param, type_params, output, ', ')
        output.append('>')

@formatter
def output_declaration(declaration, output):
    for annotation in declaration.annotations:
        output_annotation(annotation, output)

    output_modifiers(declaration.modifiers, output)
    output.append(' ')

    if isinstance(declaration, javalang.tree.ClassDeclaration):
        output.append('class ')
    elif isinstance(declaration, javalang.tree.EnumDeclaration):
        output.append('enum ')
    elif isinstance(declaration, javalang.tree.InterfaceDeclaration):
        output.append('interface ')
    elif isinstance(declaration, javalang.tree.AnnotationDeclaration):
        output.append('@interface ')

    output.append(declaration.name)

    if isinstance(declaration, (javalang.tree.ClassDeclaration, javalang.tree.InterfaceDeclaration)):
        output_type_params(declaration.type_parameters, output)

    if isinstance(declaration, javalang.tree.ClassDeclaration) and declaration.extends:
        output.append(' extends ')
        output_type(declaration.extends, output)

    if isinstance(declaration, javalang.tree.InterfaceDeclaration) and declaration.extends:
        output.append(' extends ')
        output_list(output_type, declaration.extends, output, ', ')

    if isinstance(declaration, (javalang.tree.ClassDeclaration, javalang.tree.EnumDeclaration)) and declaration.implements:
        output.append(' implements ')
        output_list(output_type, declaration.implements, output, ', ')

@formatter
def output_formal_param(param, output):
    output_type(param.type, output)

    if param.varargs:
        output.append('...')

    output.append(' ')
    output.append(param.name)

@formatter
def output_modifiers(modifiers, output):
    ordered_modifiers = [mod for mod in __modifiers_order if mod in modifiers]
    output_list(lambda mod, output: output.append(mod), ordered_modifiers, output, ' ')



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: MBI

  Description: @{shortdesc}@
    @{longdesc}@

  Version of MPI: Conforms to MPI 2, requires MPI 3 implementation (for lock_all/unlock_all epochs)

BEGIN_MPI_FEATURES
  P2P!basic: Lacking
  P2P!nonblocking: Lacking
  P2P!persistent: Lacking
  COLL!basic: Lacking
  COLL!nonblocking: Lacking
  COLL!persistent: Lacking
  COLL!tools: Lacking
  RMA: @{rmafeature}@
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

#define N 1

int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;
  MPI_Win win;
  int winbuf[100] = {0};

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug!\\n");

  MPI_Datatype type = MPI_INT;
  int target = 1;

  MPI_Win_create(&winbuf, 100 * sizeof(int), sizeof(int), MPI_INFO_NULL, MPI_COMM_WORLD, &win);

  winbuf[0] = 12345;
  @{init1}@

  @{epoch}@

  if (rank == 0) {
    @{operation1}@ /* MBIERROR1 */
    @{operation2}@ /* MBIERROR2 */
  }

  @{finEpoch}@

  MPI_Win_free(&win);

  MPI_Finalize();
  return 0;
}
"""


for e in gen.epoch:
    for p1 in gen.get:
        for p2 in gen.put + gen.store + gen.load + gen.get + gen.loadstore:
            patterns = {}
            patterns = {'e': e, 'p1': p1, 'p2': p2}
            patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
            patterns['rmafeature'] = 'Yes'
            patterns['p1'] = p1
            patterns['p2'] = p2
            patterns['e'] = e
            patterns['epoch'] = gen.epoch[e]("1")
            patterns['finEpoch'] = gen.finEpoch[e]("1")
            patterns['init1'] = gen.init[p1]("1")
            patterns['operation1'] = gen.operation[p1]("1")
            patterns['operation2'] = gen.operation[p2]("1")
            shortdesc_rma = 'Correct code using RMA operations'

            # Generate a data race (Get + Get/load/store/Put)
            replace = patterns.copy()
            replace['shortdesc'] = 'Local Concurrency error.'
            replace['longdesc'] = 'Local Concurrency error. @{p2}@ conflicts with @{p1}@'
            replace['outcome'] = 'ERROR: LocalConcurrency'
            replace['errormsg'] = 'Local Concurrency error. @{p2}@ at @{filename}@:@{line:MBIERROR2}@ conflicts with @{p1}@ line @{line:MBIERROR1}@'
            gen.make_file(template, f'LocalConcurrency_lloutwindow_{e}_{p1}_{p2}_nok.c', replace)
            # Generate a correct code by switching operation1 and  operation2
            if p2 in gen.store + gen.load + gen.loadstore:
                replace = patterns.copy()
                replace['shortdesc'] = shortdesc_rma
                replace['longdesc'] = shortdesc_rma
                replace['outcome'] = 'OK'
                replace['errormsg'] = 'OK'
                replace['operation1'] = gen.operation[p2]("1")
                replace['operation2'] = gen.operation[p1]("1")
                gen.make_file(template, f'LocalConcurrency_lloutwindow_{e}_{p2}_{p1}_ok.c', replace)
        # Generate a correct code by removing operation2
        replace = patterns.copy()
        replace['shortdesc'] = shortdesc_rma
        replace['longdesc'] = shortdesc_rma
        replace['outcome'] = 'OK'
        replace['errormsg'] = 'OK'
        replace['operation1'] = gen.operation[p1]("1")
        replace['operation2'] = ''
        gen.make_file(template, f'LocalConcurrency_{e}_{p1}_ok.c', replace)


for e in gen.epoch:
    for p1 in gen.put:
        for p2 in gen.store:
            patterns = {}
            patterns = {'e': e, 'p1': p1, 'p2': p2}
            patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
            patterns['rmafeature'] = 'Yes'
            patterns['p1'] = p1
            patterns['p2'] = p2
            patterns['e'] = e
            patterns['epoch'] = gen.epoch[e]("1")
            patterns['finEpoch'] = gen.finEpoch[e]("1")
            patterns['init1'] = gen.init[p1]("1")
            patterns['operation1'] = gen.operation[p1]("1")
            patterns['operation2'] = gen.operation[p2]("1")

            # Generate a data race (Put + store)
            replace = patterns.copy()
            replace['shortdesc'] = 'Local Concurrency error.'
            replace['longdesc'] = 'Local Concurrency error. @{p2}@ conflicts with @{p1}@'
            replace['outcome'] = 'ERROR: LocalConcurrency'
            replace['errormsg'] = 'Local Concurrency error. @{p2}@ at @{filename}@:@{line:MBIERROR2}@ conflicts with @{p1}@ line @{line:MBIERROR1}@'
            gen.make_file(template, f'LocalConcurrency_lloutwindow_{e}_{p1}_{p2}_nok.c', replace)
            # Generate a correct code by switching operation1 and operation2
            replace = patterns.copy()
            replace['shortdesc'] = shortdesc_rma
            replace['longdesc'] = shortdesc_rma
            replace['outcome'] = 'OK'
            replace['errormsg'] = 'OK'
            replace['operation1'] = gen.operation[p2]("1")
            replace['operation2'] = gen.operation[p1]("1")
            gen.make_file(template, f'LocalConcurrency_lloutwindow_{e}_{p2}_{p1}_ok.c', replace)

            # Generate a correct code by removing operation2
            replace = patterns.copy()
            replace['shortdesc'] = shortdesc_rma
            replace['longdesc'] = shortdesc_rma
            replace['outcome'] = 'OK'
            replace['errormsg'] = 'OK'
            replace['operation1'] = gen.operation[p1]("1")
            replace['operation2'] = ''
            gen.make_file(template, f'LocalConcurrency_{e}_{p1}_ok.c', replace)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2019-2022. The SimGrid Team. All rights reserved.

# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
Search for symbols documented in both the XML files produced by Doxygen and the python modules,
but not documented with breathe in the RST files.

This script is tailored to SimGrid own needs.

If you are missing some dependencies, try:  pip3 install --requirement docs/requirements.txt
"""

import os
import re
import sys
import xml.etree.ElementTree as ET
import inspect

xml_files = [
    'build/xml/classsimgrid_1_1s4u_1_1Activity.xml',
    'build/xml/classsimgrid_1_1s4u_1_1Actor.xml',
    'build/xml/classsimgrid_1_1s4u_1_1Barrier.xml',
    'build/xml/classsimgrid_1_1s4u_1_1Comm.xml',
    'build/xml/classsimgrid_1_1s4u_1_1ConditionVariable.xml',
    'build/xml/classsimgrid_1_1s4u_1_1Disk.xml',
    'build/xml/classsimgrid_1_1s4u_1_1Engine.xml',
    'build/xml/classsimgrid_1_1s4u_1_1Exec.xml',
    'build/xml/classsimgrid_1_1s4u_1_1Host.xml',
    'build/xml/classsimgrid_1_1s4u_1_1Io.xml',
    'build/xml/classsimgrid_1_1s4u_1_1Link.xml',
    'build/xml/classsimgrid_1_1s4u_1_1Mailbox.xml',
    'build/xml/classsimgrid_1_1s4u_1_1Mutex.xml',
    'build/xml/classsimgrid_1_1s4u_1_1NetZone.xml',
    'build/xml/classsimgrid_1_1s4u_1_1Semaphore.xml',
    'build/xml/classsimgrid_1_1s4u_1_1VirtualMachine.xml',
    'build/xml/classsimgrid_1_1xbt_1_1signal_3_01R_07P_8_8_8_08_4.xml',
    'build/xml/namespacesimgrid_1_1s4u_1_1this__actor.xml',
    'build/xml/actor_8h.xml',
    'build/xml/barrier_8h.xml',
    'build/xml/cond_8h.xml',
    'build/xml/engine_8h.xml',
    'build/xml/forward_8h.xml',
    'build/xml/host_8h.xml',
    'build/xml/link_8h.xml',
    'build/xml/mailbox_8h.xml',
    'build/xml/msg_8h.xml',
    'build/xml/mutex_8h.xml',
    'build/xml/semaphore_8h.xml',
    'build/xml/vm_8h.xml',
    'build/xml/zone_8h.xml'
]

python_modules = [
    'simgrid'
]
python_ignore = [
    'simgrid.ActorKilled'
]


############ Search the python elements in the source, and report undocumented ones
############

# data structure in which we store the declaration we find
python_decl = {}

def handle_python_module(fullname, englobing, elm):
    """Recursive function exploring the python modules."""

    def found_decl(kind, obj):
        """Helper function that add an object in the python_decl data structure"""
        if kind not in python_decl:
            python_decl[kind] = []
        python_decl[kind].append(obj)


    if fullname in python_ignore:
        print("Ignore Python symbol '{}' as requested.".format(fullname))
        return

    if inspect.isroutine(elm) and inspect.isclass(englobing):
        found_decl("method", fullname)
#        print('.. automethod:: {}'.format(fullname))
    elif inspect.isroutine(elm) and (not inspect.isclass(englobing)):
        found_decl("function", fullname)
#        print('.. autofunction:: {}'.format(fullname))
    elif inspect.isdatadescriptor(elm):
        found_decl("attribute", fullname)
#        print('.. autoattribute:: {}'.format(fullname))
    elif isinstance(elm, (int, str)): # We do have such a data, directly in the SimGrid top module
        found_decl("data", fullname)
#        print('.. autodata:: {}'.format(fullname))
    elif inspect.ismodule(elm) or inspect.isclass(elm):
        for name, data in inspect.getmembers(elm):
            if name.startswith('__'):
                continue
#            print("Recurse on {}.{}".format(fullname, name))
            handle_python_module("{}.{}".format(fullname, name), elm, data)
    else:
        print('UNHANDLED TYPE {} : {!r} Type: {}'.format(fullname, elm, type(elm)))

# Start the recursion on the provided Python modules
for name in python_modules:
    try:
        module = __import__(name)
    except Exception:
        if os.path.exists("../lib") and "../lib" not in sys.path:
            print("Adding ../lib to PYTHONPATH as {} cannot be imported".format(name))
            sys.path.append("../lib")
            try:
                module = __import__(name)
            except Exception:
                print("Cannot import {}, even with PYTHONPATH=../lib".format(name))
                sys.exit(1)
        else:
            print("Cannot import {}".format(name))
            sys.exit(1)
    for sub in dir(module):
        if sub[0] == '_':
            continue
        handle_python_module("{}.{}".format(name, sub), module, getattr(module, sub))

# Forget about the python declarations that were actually done in the RST
for kind in python_decl:
    with os.popen('grep \'[[:blank:]]*auto{}::\' source/*rst|sed \'s/^.*auto{}:: //\''.format(kind, kind)) as pse:
        for fullname in (l.strip() for l in pse):
            if fullname not in python_decl[kind]:
                print("Warning: {} documented but declaration not found in python.".format(fullname))
            else:
                python_decl[kind].remove(fullname)
# Dump the missing ones
for kind in python_decl:
    for fullname in python_decl[kind]:
        print(" .. auto{}:: {}".format(kind, fullname))

################ And now deal with Doxygen declarations
################

doxy_funs = {} # {classname: {func_name: [args]} }
doxy_vars = {} # {classname: [names]}
doxy_type = {} # {classname: [names]}

# find the declarations in the XML files
for arg in xml_files:
    if arg[-4:] != '.xml':
        print("Argument '{}' does not end with '.xml'".format(arg))
        continue
    #print("Parse file {}".format(arg))
    tree = ET.parse(arg)
    for elem in tree.findall(".//compounddef"):
        if elem.attrib["kind"] == "class":
            if elem.attrib["prot"] != "public":
                continue
            if "compoundname" in elem:
                raise Exception("Compound {} has no 'compoundname' child tag.".format(elem))
            compoundname = elem.find("compoundname").text
            #print("compoundname {}".format(compoundname))
        elif elem.attrib["kind"] == "file":
            compoundname = ""
        elif elem.attrib["kind"] == "namespace":
            compoundname = elem.find("compoundname").text
        else:
            print("Element {} is of kind {}".format(elem.attrib["id"], elem.attrib["kind"]))

        for member in elem.findall('.//memberdef'):
            if member.attrib["prot"] != "public":
                continue
            kind = member.attrib["kind"]
            name = member.find("name").text
            #print("kind:{} compoundname:{} name:{}".format( kind,compoundname, name))
            if kind == "variable":
                if compoundname not in doxy_vars:
                    doxy_vars[compoundname] = []
                doxy_vars[compoundname].append(name)
            elif kind == "function":
                args = member.find('argsstring').text
                args = re.sub(r'\)[^)]*$', ')', args) # ignore what's after the parameters (eg, '=0' or ' const')

                if compoundname not in doxy_funs:
                    doxy_funs[compoundname] = {}
                if name not in doxy_funs[compoundname]:
                    doxy_funs[compoundname][name] = []
                doxy_funs[compoundname][name].append(args)
            elif kind == "typedef":
                if compoundname not in doxy_type:
                    doxy_type[compoundname] = []
                doxy_type[compoundname].append(name)
            elif kind == "friend":
                pass # Ignore friendship
            else:
                print("member {}::{} is of kind {}".format(compoundname, name, kind))

# Forget about the declarations that are done in the RST
with os.popen('grep doxygenfunction:: find-missing.ignore source/*rst|sed \'s/^.*doxygenfunction:: //\'|sed \'s/ *const//\'') as pse:
    for line in (l.strip() for l in pse):
        (klass, obj, args) = (None, None, None)
        if "(" in line:
            (line, args) = line.split('(', 1)
            args = "({}".format(args)
        if '::' in line:
            (klass, obj) = line.rsplit('::', 1)
        else:
            (klass, obj) = ("", line)

        if klass not in doxy_funs:
            print("Warning: {} documented, but class {} not found in doxygen.".format(line, klass))
            continue
        if obj not in doxy_funs[klass]:
            print("Warning: Object '{}' documented but not found in '{}'".format(line, klass))
#            for obj in doxy_funs[klass]:
#                print("  found: {}::{}".format(klass, obj))
        elif len(doxy_funs[klass][obj]) == 1:
            del doxy_funs[klass][obj]
        elif args not in doxy_funs[klass][obj]:
            print("Warning: Function {}{} not found in {}".format(obj, args, klass))
        else:
            #print("Found {} in {}".format(line, klass))
            doxy_funs[klass][obj].remove(args)
            if len(doxy_funs[klass][obj]) == 0:
                del doxy_funs[klass][obj]
with os.popen('grep doxygenvariable:: find-missing.ignore source/*rst|sed \'s/^.*doxygenvariable:: //\'') as pse:
    for line in (l.strip() for l in pse):
        (klass, var) = line.rsplit('::', 1)

        if klass not in doxy_vars:
            print("Warning: {} documented, but class {} not found in doxygen.".format(line, klass))
            continue
        if var not in doxy_vars[klass]:
            print("Warning: Object {} documented but not found in '{}'".format(line, klass))
        else:
#            print("Found {} in {}".format(line, klass))
            doxy_vars[klass].remove(var)
            if len(doxy_vars[klass]) == 0:
                del doxy_vars[klass]
with os.popen('grep doxygentypedef:: find-missing.ignore source/*rst|sed \'s/^.*doxygentypedef:: //\'') as pse:
    for line in (l.strip() for l in pse):
        if '::' in line:
            (klass, typ) = line.rsplit('::', 1)
        else:
            (klass, typ) = ('', line)

        if klass not in doxy_type:
            print("Warning: {} documented, but class {} not found in doxygen.".format(line, klass))
            continue
        if typ not in doxy_type[klass]:
            print("Warning: Type {} documented but not found in '{}'".format(line, klass))
        else:
#            print("Found {} in {}".format(line, klass))
            doxy_type[klass].remove(typ)
            if len(doxy_type[klass]) == 0:
                del doxy_type[klass]

# Dump the undocumented Doxygen declarations
for obj in sorted(doxy_funs):
    for meth in sorted(doxy_funs[obj]):
        for args in sorted(doxy_funs[obj][meth]):
            if obj == '':
                print(".. doxygenfunction:: {}{}".format(meth, args))
            else:
                print(".. doxygenfunction:: {}::{}{}".format(obj, meth, args))

for obj in doxy_vars:
    for meth in sorted(doxy_vars[obj]):
        print(".. doxygenvariable:: {}::{}".format(obj, meth))

for obj in doxy_type:
    for meth in sorted(doxy_type[obj]):
        if obj == '':
            print(".. doxygentypedef:: {}".format(meth))
        else:
            print(".. doxygentypedef:: {}::{}".format(obj, meth))



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from typing import List, Optional
from dataclasses import dataclass
from argparse import ArgumentParser
import sys

from simgrid import Actor, Engine, Host, Semaphore, this_actor


@dataclass
class Shared:
    buffer: str


END_MARKER = ""


shared = Shared("")
sem_empty = Semaphore(1)  # indicates whether the buffer is empty
sem_full = Semaphore(0)  # indicates whether the buffer is full


def create_parser():
    parser = ArgumentParser()
    parser.add_argument(
        '--platform',
        type=str,
        required=True,
        help='path to the platform description'
    )
    parser.add_argument(
        '--words',
        type=lambda raw_words: raw_words.split(","),
        default=["one", "two", "three"],
        help='Comma-delimited list of words sent by the producer to the consumer'
    )
    return parser


def producer(words: List[str]):
    this_actor.info("starting consuming")
    for word in words + [END_MARKER]:
        sem_empty.acquire()
        this_actor.sleep_for(1)
        this_actor.info(f"Pushing '{word}'")
        shared.buffer = word
        sem_full.release()
    this_actor.info("Bye!")


def consumer():
    this_actor.info("starting producing")
    word: Optional[str] = None
    while word != END_MARKER:
        sem_full.acquire()
        word = str(shared.buffer)
        sem_empty.release()
        this_actor.info(f"Receiving '{word}'")
    this_actor.info("Bye!")


def main():
    settings = create_parser().parse_known_args()[0]
    e = Engine(sys.argv)
    e.load_platform(settings.platform)
    Actor.create("producer", Host.by_name("Tremblay"), producer, settings.words)
    Actor.create("consumer", Host.by_name("Jupiter"), consumer)
    e.run()


if __name__ == "__main__":
    main()



#
# Copyright 2012-2015 Bronto Software, Inc. and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import unicode_literals
from builtins import str

import logging
import re
import sys

class StringBuilder(list):
    def build(self):
        return str(self)

    def __str__(self):
        return ''.join(self)

class Directive(object):

    def __init__(self, typ, argument=''):
        self.type = typ
        self.argument = argument

        self.options = []
        self.content = []

    def add_option(self, name, value=''):
        self.options.append((name, value))

    def add_content(self, o):
        assert o is not None
        self.content.append(o)

    def build(self):
        doc = Document()
        doc.add_line('.. %s:: %s' % (self.type, self.argument))

        for name, value in self.options:
            doc.add_line('   :%s: %s\n' % (name, value))

        content = Document()

        for obj in self.content:
            content.add_object(obj)

        doc.clear()
        for line in content.build().splitlines():
            doc.add_line('   ' + line)
        doc.clear()

        return doc.build()

class Document(object):
    remove_trailing_whitespace_re = re.compile('[ \t]+$', re.MULTILINE)
    collapse_empty_lines_re = re.compile('\n' + '{3,}', re.DOTALL)

    def __init__(self):
        self.content = []

    def add_object(self, o):
        assert o is not None

        self.content.append(o)

    def add(self, s):
        self.add_object(s)

    def add_line(self, s):
        self.add(s)
        self.add('\n')

    def add_heading(self, s, t='-'):
        self.add_line(s)
        self.add_line(t * len(s))

    def clear(self):
        self.add('\n\n')

    def build(self):
        output = StringBuilder()

        for obj in self.content:
            if isinstance(obj, Directive):
                output.append('\n\n')
                output.append(obj.build())
                output.append('\n\n')
            elif isinstance(obj, Document):
                output.append(obj.build())
            else:
                output.append(str(obj))

        output.append('\n\n')

        output = str(output)
        output = self.remove_trailing_whitespace_re.sub('', output)
        output = self.collapse_empty_lines_re.sub('\n\n', output)

        return output

def error(s, *args, **kwargs):
    logging.error(s, *args, **kwargs)
    sys.exit(1)

def unexpected(s, *args, **kwargs):
    logging.exception(s, *args, **kwargs)
    sys.exit(1)



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: MBI

  Description: @{shortdesc}@
    @{longdesc}@

   Version of MPI: Conforms to MPI 1.1, does not require MPI 2 implementation

BEGIN_MPI_FEATURES
  P2P!basic: Lacking
  P2P!nonblocking: @{ip2pfeature}@
  P2P!persistent: @{persfeature}@
  COLL!basic: Lacking
  COLL!nonblocking: @{icollfeature}@
  COLL!persistent: @{cpersfeature}@
  COLL!tools: Lacking
  RMA: Lacking
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>


int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;
  int root = 0;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug!\\n");

  MPI_Comm newcom = MPI_COMM_WORLD;
  MPI_Datatype type = MPI_INT;
  MPI_Op op = MPI_SUM;
  int stag = 0, rtag = 0;
  int buff_size = 1;

  int dbs = sizeof(int)*nprocs; /* Size of the dynamic buffers for alltoall and friends */

  int dest = (rank == nprocs - 1) ? (0) : (rank + 1);
  int src = (rank == 0) ? (nprocs - 1) : (rank - 1);

  @{init1}@
  @{init2}@

  @{operation1}@ /* MBIERROR */
  @{start1}@
  @{operation2}@
  @{start2}@

  @{fini1}@
  @{fini2}@

  @{free1}@
  @{free2}@

  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""


for s in gen.isend + gen.psend:
    for r in gen.irecv + gen.precv:
        patterns = {}
        patterns = {'s': s, 'r': r}
        patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
        patterns['persfeature'] = 'Yes' if s in gen.psend or r in gen.precv  else 'Lacking'
        patterns['ip2pfeature'] = 'Yes' if s in gen.isend or r in gen.irecv  else 'Lacking'
        patterns['icollfeature'] = 'Lacking'
        patterns['cpersfeature'] = 'Lacking'
        patterns['s'] = s
        patterns['r'] = r
        patterns['init1'] = gen.init[s]("1")
        patterns['init2'] = gen.init[r]("2")
        patterns['start1'] = gen.start[s]("1")
        startPers = patterns['start1']
        patterns['start2'] = gen.start[r]("2")
        patterns['operation1'] = gen.operation[s]("1")
        patterns['operation2'] = gen.operation[r]("2")
        patterns['fini1'] = gen.fini[s]("1")
        wait = patterns['fini1']
        patterns['fini2'] = gen.fini[r]("2")
        patterns['free1'] = gen.free[s]("1")
        Reqfree = patterns['free1']
        patterns['free2'] = gen.free[r]("2")

        # Generate the correct code
        replace = patterns.copy()
        replace['shortdesc'] = 'Correct matching'
        replace['longdesc'] = 'No error'
        replace['outcome'] = 'OK'
        replace['errormsg'] = 'OK'
        gen.make_file(template, f'ReqLifecycle_{s}_{r}_ok.c', replace)

        # Generate the code with a missing wait
        replace = patterns.copy()
        replace['shortdesc'] = 'Missing wait'
        replace['longdesc'] = 'Missing Wait. @{s}@ at @{filename}@:@{line:MBIERROR}@ has no completion.'
        replace['outcome'] = 'ERROR: MissingWait'
        replace['errormsg'] = 'ERROR: MissingWait'
        replace['fini1'] = ' /* MBIERROR MISSING: ' + wait + ' */'
        gen.make_file(template, f'ReqLifecycle_MissingWait_{s}_{r}_nok.c', replace)

        if s in gen.psend:
            # Generate the code with a missing start - persistent only
            replace = patterns.copy()
            replace['shortdesc'] = 'Missing start'
            replace['longdesc'] = 'Missing start. @{s}@ at @{filename}@:@{line:MBIERROR}@ has no start'
            replace['outcome'] = 'ERROR: MissingStart'
            replace['errormsg'] = 'ERROR: MissingStart'
            replace['fini1'] = gen.fini[s]("1")
            replace['start1'] = ' /* MBIERROR MISSING: ' + startPers + ' */'
            gen.make_file(template, f'ReqLifecycle_MissingStart_{s}_{r}_nok.c', replace)
            # Generate the code with a missing free - persistent only
            replace = patterns.copy()
            replace['shortdesc'] = 'Missing free'
            replace['longdesc'] = 'Missing free. @{s}@ at @{filename}@:@{line:MBIERROR}@ has no free'
            replace['outcome'] = 'ERROR: RequestLeak'
            replace['errormsg'] = 'ERROR: RequestLeak'
            replace['start1'] = gen.start[s]("1")
            replace['free1'] = ' /* MBIERROR MISSING: ' + Reqfree + ' */'
            gen.make_file(template, f'ResLeak_nofree_{s}_{r}_nok.c', replace)


# Collectives only
for c in gen.pcoll + gen.icoll + gen.ibarrier:
    patterns = {}
    patterns = {'c': c}
    patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
    patterns['persfeature'] = 'Lacking'
    patterns['ip2pfeature'] = 'Lacking'
    patterns['cpersfeature'] = 'Yes' if c in gen.pcoll else 'Lacking'
    patterns['icollfeature'] = 'Yes' if c in gen.icoll + gen.ibarrier else 'Lacking'
    patterns['c'] = c
    patterns['init1'] = gen.init[c]("1")
    patterns['operation1'] = gen.operation[c]("1")
    patterns['start1'] = gen.start[c]("1")
    patterns['fini1'] = gen.fini[c]("1")
    patterns['free1'] = gen.free[c]("1")
    opstart = patterns['start1']
    opwait = patterns['fini1']
    opfree = patterns['free1']
    patterns['init2'] = ""
    patterns['operation2'] = ""
    patterns['start2'] = ""
    patterns['fini2'] = ""
    patterns['free2'] = ""

    # Generate the code with a missing wait
    replace = patterns.copy()
    replace['shortdesc'] = 'Missing wait'
    replace['longdesc'] = 'Missing Wait. @{c}@ at @{filename}@:@{line:MBIERROR}@ has no completion'
    replace['outcome'] = 'ERROR: MissingWait'
    replace['errormsg'] = 'ERROR: MissingWait'
    replace['fini1'] = ' /* MBIERROR MISSING: ' + opwait + ' */'
    replace['free1'] = ' /* MISSING: ' + replace['free1'] + ' (to not free the buffer before an internal wait */'
    gen.make_file(template, f'ReqLifecycle_MissingWait_{c}_nok.c', replace)

    if c in gen.pcoll:
        # Generate the code with a missing start - persistent only
        replace = patterns.copy()
        replace['shortdesc'] = 'Missing start functio'
        replace['longdesc'] = 'Missing Start. @{c}@ at @{filename}@:@{line:MBIERROR}@ has no start'
        replace['outcome'] = 'ERROR: MissingStart'
        replace['errormsg'] = 'ERROR: MissingStart'
        replace['fini1'] = gen.fini[c]("1")
        replace['start1'] = ' /* MBIERROR MISSING: ' + opstart + ' */'
        gen.make_file(template, f'ReqLifecycle_MissingStart_{c}_nok.c', replace)

        # Generate the code with a resleak (no free) - persistent only
        replace = patterns.copy()
        replace['shortdesc'] = 'Missing free'
        replace['longdesc'] = 'Missing free. @{c}@ at @{filename}@:@{line:MBIERROR}@ has no free'
        replace['outcome'] = 'ERROR: RequestLeak'
        replace['errormsg'] = 'ERROR: RequestLeak'
        replace['start1'] = gen.start[c]("1")
        replace['free1'] = ' /* MBIERROR MISSING: ' + opfree + ' */'
        gen.make_file(template, f'ResLeak_nofree_{c}_nok.c', replace)



# Copyright (c) 2018-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

import sys
from simgrid import Actor, Engine, Host, this_actor


def executor():
    # execute() tells SimGrid to pause the calling actor until
    # its host has computed the amount of flops passed as a parameter
    this_actor.execute(98095)
    this_actor.info("Done.")
    # This simple example does not do anything beyond that


def privileged():
    # You can also specify the priority of your execution as follows.
    # An execution of priority 2 computes twice as fast as a regular one.
    #
    # So instead of a half/half sharing between the two executions,
    # we get a 1/3 vs 2/3 sharing.
    this_actor.execute(98095, priority=2)
    this_actor.info("Done.")

    # Note that the timings printed when executing this example are a bit misleading,
    # because the uneven sharing only last until the privileged actor ends.
    # After this point, the unprivileged one gets 100% of the CPU and finishes
    # quite quickly.


if __name__ == '__main__':
    e = Engine(sys.argv)
    e.load_platform(sys.argv[1])

    Actor.create("executor", Host.by_name("Tremblay"), executor)
    Actor.create("privileged", Host.by_name("Tremblay"), privileged)

    e.run()



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: @{origin}@

  Description: @{shortdesc}@
    @{longdesc}@

    Version of MPI: Conforms to MPI 1.1, does not require MPI 2 implementation

BEGIN_MPI_FEATURES
    P2P!basic: Lacking
    P2P!nonblocking: Lacking
    P2P!persistent: Lacking
    COLL!basic: Lacking
    COLL!nonblocking: Lacking
    COLL!persistent: Lacking
    COLL!tools: Lacking
    RMA: @{rmafeature}@
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

#define N 20

int main(int argc, char **argv) {
  int rank, numProcs;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &numProcs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  if (numProcs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug!\\n");

  int *winbuf = malloc(N * sizeof(int));

  MPI_Win win;
  MPI_Win_create(winbuf, N * sizeof(int), 1, MPI_INFO_NULL, MPI_COMM_WORLD, &win);

  MPI_Datatype type = MPI_INT;
  int target = 1;

  @{epoch}@

  if (rank == 0) {
    @{epoch2}@

    @{init}@
    @{operation}@

    @{finEpoch2}@
  }

  @{finEpoch}@

  MPI_Win_free(&win);

  free(winbuf);

  MPI_Finalize();

  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""


for e1 in gen.epoch:
    for p in gen.rma:
        patterns = {}
        patterns = {'e1': e1, 'p': p}
        patterns['origin'] = "MPI-Corrbench"
        patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
        patterns['rmafeature'] = 'Yes'
        patterns['p'] = p
        patterns['e1'] = e1
        patterns['epoch'] = gen.epoch[e1]("1")
        patterns['finEpoch'] = gen.finEpoch[e1]("1")
        patterns['epoch2'] = ""
        patterns['finEpoch2'] = ""
        patterns['init'] = gen.init[p]("1")
        patterns['operation'] = gen.operation[p]("1")

        # Generate a code correct
        replace = patterns.copy()
        replace['shortdesc'] = 'Correct code'
        replace['longdesc'] = 'Correct code'
        replace['outcome'] = 'OK'
        replace['errormsg'] = 'OK'
        gen.make_file(template, f'ReqLifecycle_RMA_{e1}_{p}_ok.c', replace)

        # Generate a code with missing open epoch
        replace = patterns.copy()
        replace['shortdesc'] = f"Request lifecycle, missing open {e1} epoch"
        replace['longdesc'] = f"Request lifecycle, missing open {e1} epoch"
        replace['outcome'] = 'ERROR: MissingStart'
        replace['errormsg'] = '@{e1}@ at @{filename}@:@{line:MBIERROR}@ has missing'
        replace['epoch'] = f"/* MBIERROR MISSING: {gen.epoch[e1]('1')} */"
        gen.make_file(template, f'ReqLifecycle_RMA_MissingOpen_{e1}_{p}_nok.c', replace)

        # Generate a code with missing close epoch
        replace = patterns.copy()
        replace['shortdesc'] = f"Request lifecycle, missing close {e1} epoch"
        replace['longdesc'] = f"Request lifecycle, missing close {e1} epoch"
        replace['outcome'] = 'ERROR: MissingWait'
        replace['errormsg'] = '@{e1}@ at @{filename}@:@{line:MBIERROR}@ has missing'
        replace['epoch'] = gen.epoch[e1]("1")
        replace['finEpoch'] = f"/* MBIERROR MISSING: {gen.finEpoch[e1]('1')} */"
        gen.make_file(template, f'ReqLifecycle_RMA_MissingClose_{e1}_{p}_nok.c', replace)

for e1 in gen.epoch:
    for e2 in gen.epoch:
        for p in gen.rma:
            patterns = {}
            patterns = {'e1': e1, 'e2': e2, 'p': p}
            patterns['origin'] = "MPI-Corrbench"
            patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
            patterns['rmafeature'] = 'Yes'
            patterns['p'] = p
            patterns['e1'] = e1
            patterns['e2'] = e2
            patterns['epoch'] = gen.epoch[e1]("1")
            patterns['finEpoch'] = gen.finEpoch[e1]("1")
            patterns['epoch2'] = gen.epoch[e2]("1") + " /* MBIERROR */"
            patterns['finEpoch2'] = gen.finEpoch[e2]("1") + " /* MBIERROR */"
            patterns['init'] = gen.init[p]("1")
            patterns['operation'] = gen.operation[p]("1")

            # Generate a code with epoch into an epoch
            replace = patterns.copy()
            replace['shortdesc'] = f"Request lifecycle, {e2} epoch into {e1} epoch"
            replace['longdesc'] = f"Request lifecycle, {e2} epoch into {e1} epoch"
            replace['outcome'] = 'ERROR: MissingWait' #FIXME: New type of error
            replace['errormsg'] = '@{e2}@ at @{filename}@:@{line:MBIERROR}@ has in an other epoch'
            gen.make_file(template, f'ReqLifecycle_RMA_TwoEpoch_{e1}_{e2}_{p}_nok.c', replace)



#! /usr/bin/python3

# Copyright 2021-2022. The MBI project. All rights reserved.
# This program is free software; you can redistribute it and/or modify it under the terms of the license (GNU GPL).

import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: MBI

  Description: @{shortdesc}@
    @{longdesc}@

  Version of MPI: Conforms to MPI 1.1, does not require MPI 2 implementation

BEGIN_MPI_FEATURES
  P2P!basic: Lacking
  P2P!nonblocking: Lacking
  P2P!persistent: Lacking
  COLL!basic: Lacking
  COLL!nonblocking: Lacking
  COLL!persistent: Lacking
  COLL!tools: @{toolfeature}@
  RMA: Lacking
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>

#define ITERATIONS 100
#define PARAM_PER_ITERATION 3
#define PARAM_LOST_PER_ITERATION 1

static void myOp(int *invec, int *inoutvec, int *len, MPI_Datatype *dtype) {
  for (int i = 0; i < *len; i++)
    inoutvec[i] += invec[i];
}

int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;
  int size = 1;
  int j = 0;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug!\\n");

  @{change_size}@
  @{init}@
  @{loop}@
  @{operation}@
  @{cond}@
  @{fini}@
  @{end}@

  @{free}@

  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""

# Generate code with one collective
for call in gen.tcoll:
    patterns = {}
    patterns = {'call': call}
    patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
    patterns['toolfeature'] = 'Yes'
    patterns['call'] = call
    patterns['operation'] = gen.operation[call]("1")
    patterns['init'] = gen.init[call]("1")
    patterns['fini'] = gen.fini[call]("1")
    patterns['free'] = gen.free[call]("1")
    missing = patterns['fini']
    patterns['loop'] = ''
    patterns['cond'] = ''
    patterns['change_size'] = ''
    patterns['end'] = ''

    # Generate the correct code
    replace = patterns.copy()
    replace['shortdesc'] = '@{call}@ is correctly used'
    replace['longdesc'] = f'{call} correctly used'
    replace['outcome'] = 'OK'
    replace['errormsg'] = ''
    gen.make_file(template, f'ResLeak_{call}_ok.c', replace)

    # Generate the resleak
    replace = patterns.copy()
    replace['shortdesc'] = '@{call}@ has no free'
    replace['longdesc'] = '@{call}@ has no free'
    replace['outcome'] = f'ERROR: {gen.error[call]}'
    replace['errormsg'] = 'Resleak. @{call}@ at @{filename}@:@{line:MBIERROR}@ has no free.'
    replace['fini'] = ' /* MBIERROR MISSING: ' + missing + ' */'
    gen.make_file(template, f'ResLeak_{call}_nok.c', replace)

    # Generate multiple resleak
    replace = patterns.copy()
    replace['shortdesc'] = '@{call}@ lacks several free'
    replace['longdesc'] = '@{call}@ lacks several free'
    replace['outcome'] = f'ERROR: {gen.error[call]}'
    replace['errormsg'] = 'Resleak. @{call}@ at @{filename}@:@{line:MBIERROR}@ lacks several free.'
    replace['change_size'] = 'size=PARAM_PER_ITERATION;'
    replace['loop'] = 'for (int i = 0; i < ITERATIONS; i++) {\n    for (j = 0; j < PARAM_PER_ITERATION; j++) {'
    replace['cond'] = '      if (j < PARAM_PER_ITERATION - PARAM_LOST_PER_ITERATION) {'
    replace['fini'] = gen.fini[call]("1") + ' /* MBIERROR */'
    replace['end'] = '      }\n     }\n   }'
    gen.make_file(template, f'ResLeak_multiple_{call}_nok.c', replace)



#! /usr/bin/env python3

# Copyright (c) 2006-2022. The SimGrid Team. All rights reserved.

# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.


# This script takes as input a C++ platform file, compiles it, then dumps the
# routing graph as a CSV and generates an image.
# The layout should be alright for any platform file, but the colors are very
# ad-hoc for file supernode.cpp : do not hesitate to adapt this script to your needs.
# An option is provided to "simplify" the graph by removing the link vertices. It assumes that these vertices have
# "link" in their name.

import sys
import subprocess
import pandas
import matplotlib.pyplot as plt
import networkx as nx
import argparse
import tempfile
import os

try:
    from palettable.colorbrewer.qualitative import Set1_9
    colors = Set1_9.hex_colors
except ImportError:
    print('Warning: could not import palettable, will use a default palette.')
    colors = [None]*10


def run_command(cmd):
    print(cmd)
    proc = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    _stdout, stderr = proc.communicate()
    if proc.returncode != 0:
        sys.exit(f'Command failed:\n{stderr.decode()}')


def compile_platform(platform_cpp, platform_so):
    cmd = f'g++ -g -fPIC -shared -o {platform_so} {platform_cpp} -lsimgrid'
    run_command(cmd)


def dump_csv(platform_so, platform_csv):
    cmd = f'graphicator {platform_so} {platform_csv}'
    run_command(cmd)


def merge_updown(graph):
    '''
    Merge all the UP and DOWN links.
    '''
    graph2 = graph.copy()
    downlinks = [v for v in graph if 'DOWN' in v]
    mapping = {}
    for down in downlinks:
        up = down.replace('DOWN', 'UP')
        graph2 = nx.contracted_nodes(graph2, down, up)
        mapping[down] = down.replace('_DOWN', '')
    return nx.relabel_nodes(graph2, mapping)


def contract_links(graph):
    '''
    Remove all the 'link' vertices from the graph to directly connect the nodes.
    Note: it assumes that link vertices have the "link" string in their name.
    '''
    links = [v for v in graph if 'link' in v]
    new_edges = []
    for v in links:
        neigh = [u for u in graph.neighbors(v) if 'link' not in u]  # with Floyd zones, we have links connected to links
        assert len(neigh) == 2
        new_edges.append(neigh)
    # Adding edges from graph that have no links
    for u, v in graph.edges:
        if 'link' not in u and 'link' not in v:
            new_edges.append((u, v))
    return nx.from_edgelist(new_edges)


def load_graph(platform_csv, simplify_graph):
    edges = pandas.read_csv(platform_csv)
    graph = nx.Graph()
    graph.add_edges_from([e for _, e in edges.drop_duplicates().iterrows()])
    print(f'Loaded a graph with {len(graph)} vertices with {len(graph.edges)} edges')
    if simplify_graph:
        graph = contract_links(merge_updown(graph))
        print(f'Simplified the graph, it now has {len(graph)} vertices with {len(graph.edges)} edges')
    return graph


def plot_graph(graph, label=False, groups=None):
    if groups is None:
        groups = []
    # First, we compute the graph layout, i.e. the position of the nodes.
    # The neato algorithm from graphviz is nicer, but this is an extra-dependency.
    # The spring_layout is also not too bad.
    try:
        pos = nx.nx_agraph.graphviz_layout(graph, 'neato')
    except ImportError:
        print('Warning: could not import pygraphviz, will use another layout algorithm.')
        pos = nx.spring_layout(graph, k=0.5, iterations=1000, seed=42)
    plt.figure(figsize=(20, 15))
    plt.axis('off')
    all_nodes = set(graph)
    # We then iterate on all the specified groups, to plot each of them in the right color.
    # Note that the order of the groups is important here, as we are looking at substrings in the node names.
    for i, grp in enumerate(groups):
        nodes = {u for u in all_nodes if grp in u}
        all_nodes -= nodes
        nx.draw_networkx_nodes(graph, pos, nodelist=nodes, node_size=50, node_color=colors[i], label=grp.replace('_', ''))
    nx.draw_networkx_nodes(graph, pos, nodelist=all_nodes, node_size=50, node_color=colors[-1], label='unknown')
    # Finally we draw the edges, the (optional) labels, and the legend.
    nx.draw_networkx_edges(graph, pos, alpha=0.3)
    if label:
        nx.draw_networkx_labels(graph, pos)
    plt.legend(scatterpoints=1)


def generate_svg(platform_csv, output_file, simplify_graph):
    graph = load_graph(platform_csv, simplify_graph)
    plot_graph(graph, label=False, groups=['router', 'link', 'cpu', '_node', 'supernode', 'cluster'])
    plt.savefig(output_file)
    print(f'Generated file {output_file}')


def main():
    parser = argparse.ArgumentParser(description='Visualization of topologies for SimGrid C++ platforms')
    parser.add_argument('input', type=str, help='SimGrid C++ platform file name (input)')
    parser.add_argument('output', type=str, help='File name of the output image')
    parser.add_argument('--simplify', action='store_true', help='Simplify the topology (removing link vertices)')
    args = parser.parse_args()
    if not args.input.endswith('.cpp'):
        parser.error(f'SimGrid platform must be a C++ file (with .cpp extension), got {args.input}')
    if not os.path.isfile(args.input):
        parser.error(f'File {args.input} not found')
    output_dir = os.path.dirname(args.output)
    if output_dir != '' and not os.path.isdir(output_dir):
        parser.error(f'Not a directory: {output_dir}')
    with tempfile.TemporaryDirectory() as tmpdirname:
        platform_cpp = args.input
        platform_csv = os.path.join(tmpdirname, 'platform.csv')
        platform_so = os.path.join(tmpdirname, 'platform.so')
        compile_platform(platform_cpp, platform_so)
        dump_csv(platform_so, platform_csv)
        generate_svg(platform_csv, args.output, args.simplify)


if __name__ == '__main__':
    main()



# Copyright (c) 2006-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
This example shows how to simulate a non-linear resource sharing for network links.
"""

import functools
import sys
from simgrid import Actor, Engine, Comm, Mailbox, NetZone, Link, LinkInRoute, this_actor

class Sender:
    """
    Send a series of messages to mailbox "receiver"
    """
    def __init__(self, msg_count: int, msg_size=int(1e6)):
        self.msg_count = msg_count
        self.msg_size = msg_size

    # Actors that are created as object will execute their __call__ method.
    # So, the following constitutes the main function of the Sender actor.
    def __call__(self):
        pending_comms = []
        mbox = Mailbox.by_name("receiver")

        for i in range(self.msg_count):
            msg = "Message " + str(i)
            size = self.msg_size * (i + 1)
            this_actor.info("Send '%s' to '%s, msg size: %d'" % (msg, mbox.name, size))
            comm = mbox.put_async(msg, size)
            pending_comms.append(comm)

        this_actor.info("Done dispatching all messages")

        # Now that all message exchanges were initiated, wait for their completion in one single call
        Comm.wait_all(pending_comms)

        this_actor.info("Goodbye now!")

class Receiver:
    """
    Receiver actor: wait for N messages on the mailbox "receiver"
    """

    def __init__(self, msg_count=10):
        self.msg_count = msg_count

    def __call__(self):
        mbox = Mailbox.by_name("receiver")

        pending_msgs = []
        pending_comms = []

        this_actor.info("Wait for %d messages asynchronously" % self.msg_count)
        for _ in range(self.msg_count):
            comm, data = mbox.get_async()
            pending_comms.append(comm)
            pending_msgs.append(data)

        while pending_comms:
            index = Comm.wait_any(pending_comms)
            msg = pending_msgs[index].get()
            this_actor.info("I got '%s'." % msg)
            del pending_comms[index]
            del pending_msgs[index]

####################################################################################################
def link_nonlinear(link: Link, capacity: float, n: int) -> float:
    """
    Non-linear resource sharing for links

    Note that the callback is called twice in this example:
    1) link UP: with the number of active flows (from 9 to 1)
    2) link DOWN: with 0 active flows. A crosstraffic communication is happing
    in the down link, but it's not considered as an active flow.
    """
    # emulates a degradation in link according to the number of flows
    # you probably want something more complex than that and based on real
    # experiments
    capacity = min(capacity, capacity * (1.0 - (n - 1) / 10.0))
    this_actor.info("Link %s, %d active communications, new capacity %f" % (link.name, n, capacity))
    return capacity

def load_platform():
    """
    Create a simple 2-hosts platform
     ________                 __________
    | Sender |===============| Receiver |
    |________|    Link1      |__________|

    """
    zone = NetZone.create_full_zone("Zone1")
    sender = zone.create_host("sender", 1).seal()
    receiver = zone.create_host("receiver", 1).seal()

    link = zone.create_split_duplex_link("link1", 1e6)
    # setting same callbacks (could be different) for link UP/DOWN in split-duplex link
    link.link_up.set_sharing_policy(Link.SharingPolicy.NONLINEAR,
                                    functools.partial(link_nonlinear, link.link_up))
    link.link_down.set_sharing_policy(Link.SharingPolicy.NONLINEAR,
                                      functools.partial(link_nonlinear, link.link_down))
    link.set_latency(10e-6).seal()

    # create routes between nodes
    zone.add_route(sender.netpoint, receiver.netpoint, None, None,
                   [LinkInRoute(link, LinkInRoute.Direction.UP)], True)
    zone.seal()

    # create actors Sender/Receiver
    Actor.create("receiver", receiver, Receiver(9))
    Actor.create("sender", sender, Sender(9))

###################################################################################################

if __name__ == '__main__':
    e = Engine(sys.argv)

    # create platform
    load_platform()

    # runs the simulation
    e.run()

    # explicitly deleting Engine object to avoid segfault during cleanup phase.
    # During Engine destruction, the cleanup of std::function linked to link_non_linear callback is called.
    # If we let the cleanup by itself, it fails trying on its destruction because the python main program
    # has already freed its variables
    del e



#! /usr/bin/env python3

# Copyright 2021-2022. The SimGrid Team. All rights reserved.
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

# This script runs a previously compiled MBI script and checks on the result.
# This file was written for the SimGrid project. Parts coming from the MBI projets are MBIutils.py and simgrid.py

import sys
import os

import MBIutils as mbi
import simgrid as sg

if len(sys.argv) != 4:
    print(f"Usage: MBI.py binary source (received: {sys.argv})")
    sys.exit(1)

if not os.path.exists("cluster.xml"):
    with open('cluster.xml', 'w') as outfile:
        outfile.write("<?xml version='1.0'?>\n")
        outfile.write("<!DOCTYPE platform SYSTEM \"https://simgrid.org/simgrid.dtd\">\n")
        outfile.write('<platform version="4.1">\n')
        outfile.write(' <cluster id="acme" prefix="node-" radical="0-99" suffix="" speed="1Gf" bw="125MBps" lat="50us"/>\n')
        outfile.write('</platform>\n')


simgrid = sg.Tool()

(name, path, binary, filename) = sys.argv
for test in mbi.parse_one_code(filename):
    execcmd = test['cmd'].replace("mpirun", f"{path}/smpi_script/bin/smpirun -wrapper '{path}/bin/simgrid-mc --log=mc_safety.t:info' -platform ./cluster.xml -analyze --cfg=smpi/finalization-barrier:on --cfg=smpi/list-leaks:10 --cfg=model-check/max-depth:10000")
    execcmd = execcmd.replace('${EXE}', binary)
    execcmd = execcmd.replace('$zero_buffer', "--cfg=smpi/buffering:zero")
    execcmd = execcmd.replace('$infty_buffer', "--cfg=smpi/buffering:infty")

    if os.path.exists(f'{filename}.txt'):
        os.remove(f'{filename}.txt')
    mbi.run_cmd(buildcmd="", execcmd=execcmd, cachefile=filename, filename=filename, binary=binary, timeout=300, batchinfo="", read_line_lambda=None)

    (res_category, elapsed, diagnostic, outcome) = mbi.categorize(simgrid, "simgrid", filename, test['expect'])

    if res_category != "TRUE_NEG" and res_category != "TRUE_POS":
        print("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n")
        print(f"SimGrid gave the wrong result ({outcome} instead of {test['expect']}).")
        sys.exit(1)



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from argparse import ArgumentParser
from dataclasses import dataclass
import sys

from simgrid import Actor, Engine, Host, Mutex, this_actor


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        '--platform',
        type=str,
        required=True,
        help='path to the platform description'
    )
    parser.add_argument(
        '--actors',
        type=int,
        default=6,
        help='how many pairs of actors should be started'
    )
    return parser


@dataclass
class ResultHolder:
    value: int


def worker_context_manager(mutex: Mutex, result: ResultHolder):
    # When using a context manager, the lock and the unlock are automatic. This is the easiest approach
    with mutex:
        this_actor.info(f"Hello simgrid, I'm ready to compute after acquiring the mutex from a context manager")
        result.value += 1
    this_actor.info(f"I'm done, good bye")


def worker(mutex: Mutex, result: ResultHolder):
    # If you lock your mutex manually, you also have to unlock it.
    # Beware of exceptions preventing your unlock() from being executed!
    mutex.lock()
    this_actor.info("Hello simgrid, I'm ready to compute after a regular lock")
    result.value += 1
    mutex.unlock()
    this_actor.info("I'm done, good bye")



def main():
    settings = create_parser().parse_known_args()[0]
    e = Engine(sys.argv)
    e.load_platform(settings.platform)

    # Create the requested amount of actors pairs. Each pair has a specific mutex and cell in `result`
    results = [ResultHolder(value=0) for _ in range(settings.actors)]
    for i in range(settings.actors):
        mutex = Mutex()
        Actor.create(f"worker-{i}(mgr)", Host.by_name("Jupiter"), worker_context_manager, mutex, results[i])
        Actor.create(f"worker-{i}", Host.by_name("Tremblay"), worker, mutex, results[i])

    e.run()

    for i in range(settings.actors):
        this_actor.info(f"Result[{i}] -> {results[i].value}")


if __name__ == "__main__":
    main()



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

from argparse import ArgumentParser
from typing import List
import sys

from simgrid import Actor, Comm, Engine, Mailbox, this_actor


FINALIZE_MESSAGE = "finalize"


def create_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument(
        '--platform',
        type=str,
        required=True,
        help='path to the platform description'
    )
    return parser


def get_peer_mailbox(peer_id: int) -> Mailbox:
    return Mailbox.by_name(f"peer-{peer_id}")


def peer(my_id: int, message_count: int, payload_size: int, peers_count: int):
    my_mailbox: Mailbox = get_peer_mailbox(my_id)
    my_mailbox.set_receiver(Actor.self())
    pending_comms: List[Comm] = []
    # Start dispatching all messages to peers others that myself
    for i in range(message_count):
        for peer_id in range(peers_count):
            if peer_id != my_id:
                peer_mailbox = get_peer_mailbox(peer_id)
                message = f"Message {i} from peer {my_id}"
                this_actor.info(f"Send '{message}' to '{peer_mailbox.name}'")
                pending_comms.append(peer_mailbox.put_async(message, payload_size))

    # Start sending messages to let peers know that they should stop
    for peer_id in range(peers_count):
        if peer_id != my_id:
            peer_mailbox = get_peer_mailbox(peer_id)
            payload = str(FINALIZE_MESSAGE)
            pending_comms.append(peer_mailbox.put_async(payload, payload_size))
            this_actor.info(f"Send '{payload}' to '{peer_mailbox.name}'")
    this_actor.info("Done dispatching all messages")

    # Retrieve all the messages other peers have been sending to me until I receive all the corresponding "Finalize"
    # messages
    pending_finalize_messages = peers_count - 1
    while pending_finalize_messages > 0:
        if my_mailbox.ready:
            start = Engine.clock
            received: str = my_mailbox.get()
            waiting_time = Engine.clock - start
            if waiting_time != 0.0:
                raise AssertionError(f"Expecting the waiting time to be 0.0 because the communication was supposedly "
                                     f"ready, but got {waiting_time} instead")
            this_actor.info(f"I got a '{received}'.")
            if received == FINALIZE_MESSAGE:
                pending_finalize_messages -= 1
        else:
            this_actor.info("Nothing ready to consume yet, I better sleep for a while")
            this_actor.sleep_for(0.01)

    this_actor.info("I'm done, just waiting for my peers to receive the messages before exiting")
    Comm.wait_all(pending_comms)
    this_actor.info("Goodbye now!")


def main():
    settings = create_parser().parse_known_args()[0]
    e = Engine(sys.argv)
    e.load_platform(settings.platform)
    Actor.create("peer", e.host_by_name("Tremblay"), peer, 0, 2, int(5e7), 3)
    Actor.create("peer", e.host_by_name("Ruby"), peer, 1, 6, int(2.5e5), 3)
    Actor.create("peer", e.host_by_name("Perl"), peer, 2, 0, int(5e7), 3)
    e.run()


if __name__ == "__main__":
    main()



#
# Copyright 2012-2015 Bronto Software, Inc. and contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import print_function, unicode_literals

try:
   import cPickle as pickle
except ImportError:
   import pickle

import hashlib
import logging
import sys
import os
import os.path

from optparse import OptionParser

import javalang

import javasphinx.compiler as compiler
import javasphinx.util as util

def encode_output(s):
   if isinstance(s, str):
      return s
   return s.encode('utf-8')

def find_source_files(input_path, excludes):
    """ Get a list of filenames for all Java source files within the given
    directory.

    """

    java_files = []

    input_path = os.path.normpath(os.path.abspath(input_path))

    for dirpath, dirnames, filenames in os.walk(input_path):
        if is_excluded(dirpath, excludes):
            del dirnames[:]
            continue

        for filename in filenames:
            if filename.endswith(".java"):
                java_files.append(os.path.join(dirpath, filename))

    return java_files

def write_toc(packages, opts):
    doc = util.Document()
    doc.add_heading(opts.toc_title, '=')

    toc = util.Directive('toctree')
    toc.add_option('maxdepth', '2')
    doc.add_object(toc)

    for package in sorted(packages.keys()):
        toc.add_content("%s/package-index\n" % package.replace('.', '/'))

    filename = 'packages.' + opts.suffix
    fullpath = os.path.join(opts.destdir, filename)

    if os.path.exists(fullpath) and not (opts.force or opts.update):
        sys.stderr.write(fullpath + ' already exists. Use -f to overwrite.\n')
        sys.exit(1)

    f = open(fullpath, 'w')
    f.write(encode_output(doc.build()))
    f.close()

def write_documents(packages, documents, sources, opts):
    package_contents = dict()

    # Write individual documents
    for fullname, (package, name, document) in documents.items():
        if is_package_info_doc(name):
            continue

        package_path = package.replace('.', os.sep)
        filebasename = name.replace('.', '-')
        filename = filebasename + '.' + opts.suffix
        dirpath = os.path.join(opts.destdir, package_path)
        fullpath = os.path.join(dirpath, filename)

        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        elif os.path.exists(fullpath) and not (opts.force or opts.update):
            sys.stderr.write(fullpath + ' already exists. Use -f to overwrite.\n')
            sys.exit(1)

        # Add to package indexes
        package_contents.setdefault(package, list()).append(filebasename)

        if opts.update and os.path.exists(fullpath):
            # If the destination file is newer than the source file than skip
            # writing it out
            source_mod_time = os.stat(sources[fullname]).st_mtime
            dest_mod_time = os.stat(fullpath).st_mtime

            if source_mod_time < dest_mod_time:
                continue

        f = open(fullpath, 'w')
        f.write(encode_output(document))
        f.close()

    # Write package-index for each package
    for package, classes in package_contents.items():
        doc = util.Document()
        doc.add_heading(package, '=')

        #Adds the package documentation (if any)
        if packages[package] != '':
            documentation = packages[package]
            doc.add_line("\n%s" % documentation)

        doc.add_object(util.Directive('java:package', package))

        toc = util.Directive('toctree')
        toc.add_option('maxdepth', '1')

        classes.sort()
        for filebasename in classes:
            toc.add_content(filebasename + '\n')
        doc.add_object(toc)

        package_path = package.replace('.', os.sep)
        filename = 'package-index.' + opts.suffix
        dirpath = os.path.join(opts.destdir, package_path)
        fullpath = os.path.join(dirpath, filename)

        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        elif os.path.exists(fullpath) and not (opts.force or opts.update):
            sys.stderr.write(fullpath + ' already exists. Use -f to overwrite.\n')
            sys.exit(1)

        f = open(fullpath, 'w')
        f.write(encode_output(doc.build()))
        f.close()

def get_newer(a, b):
    if not os.path.exists(a):
        return b

    if not os.path.exists(b):
        return a

    a_mtime = int(os.stat(a).st_mtime)
    b_mtime = int(os.stat(b).st_mtime)

    if a_mtime < b_mtime:
        return b

    return a

def format_syntax_error(e):
    rest = ""
    if e.at.position:
        value = e.at.value
        pos = e.at.position
        rest = ' at %s line %d, character %d' % (value, pos[0], pos[1])
    return e.description + rest

def generate_from_source_file(doc_compiler, source_file, cache_dir):
    if cache_dir:
        fingerprint = hashlib.md5(source_file.encode()).hexdigest()
        cache_file = os.path.join(cache_dir, 'parsed-' + fingerprint + '.p')

        if get_newer(source_file, cache_file) == cache_file:
            return pickle.load(open(cache_file, 'rb'))
    else:
        cache_file = None

    f = open(source_file)
    source = f.read()
    f.close()

    try:
        ast = javalang.parse.parse(source)
    except javalang.parser.JavaSyntaxError as e:
        util.error('Syntax error in %s: %s', source_file, format_syntax_error(e))
    except Exception:
        util.unexpected('Unexpected exception while parsing %s', source_file)

    documents = {}
    try:
        if source_file.endswith("package-info.java"):
            if ast.package is not None:
                documentation = doc_compiler.compile_docblock(ast.package)
                documents[ast.package.name] = (ast.package.name, 'package-info', documentation)
        else:
            documents = doc_compiler.compile(ast)
    except Exception:
        util.unexpected('Unexpected exception while compiling %s', source_file)

    if cache_file:
        dump_file = open(cache_file, 'wb')
        pickle.dump(documents, dump_file)
        dump_file.close()

    return documents

def generate_documents(source_files, cache_dir, verbose, member_headers, parser):
    documents = {}
    sources = {}
    doc_compiler = compiler.JavadocRestCompiler(None, member_headers, parser)

    for source_file in source_files:
        if verbose:
            print('Processing', source_file)

        this_file_documents = generate_from_source_file(doc_compiler, source_file, cache_dir)
        for fullname in this_file_documents:
            sources[fullname] = source_file

        documents.update(this_file_documents)

    #Existing packages dict, where each key is a package name
    #and each value is the package documentation (if any)
    packages = {}

    #Gets the name of the package where the document was declared
    #and adds it to the packages dict with no documentation.
    #Package documentation, if any, will be collected from package-info.java files.
    for package, name, _ in documents.values():
        packages[package] = ""

    #Gets packages documentation from package-info.java documents (if any).
    for package, name, content in documents.values():
        if is_package_info_doc(name):
            packages[package] = content

    return packages, documents, sources

def normalize_excludes(rootpath, excludes):
    f_excludes = []
    for exclude in excludes:
        if not os.path.isabs(exclude) and not exclude.startswith(rootpath):
            exclude = os.path.join(rootpath, exclude)
        f_excludes.append(os.path.normpath(exclude) + os.path.sep)
    return f_excludes

def is_excluded(root, excludes):
    sep = os.path.sep
    if not root.endswith(sep):
        root += sep
    for exclude in excludes:
        if root.startswith(exclude):
            return True
    return False

def is_package_info_doc(document_name):
    ''' Checks if the name of a document represents a package-info.java file. '''
    return document_name == 'package-info'


def main(argv=sys.argv):
    logging.basicConfig(level=logging.WARN)

    parser = OptionParser(
        usage="""\
usage: %prog [options] -o <output_path> <input_path> [exclude_paths, ...]

Look recursively in <input_path> for Java sources files and create reST files
for all non-private classes, organized by package under <output_path>. A package
index (package-index.<ext>) will be created for each package, and a top level
table of contents will be generated named packages.<ext>.

Paths matching any of the given exclude_paths (interpreted as regular
expressions) will be skipped.

Note: By default this script will not overwrite already created files.""")

    parser.add_option('-o', '--output-dir', action='store', dest='destdir',
                      help='Directory to place all output', default='')
    parser.add_option('-f', '--force', action='store_true', dest='force',
                      help='Overwrite all files')
    parser.add_option('-c', '--cache-dir', action='store', dest='cache_dir',
                      help='Directory to stored cachable output')
    parser.add_option('-u', '--update', action='store_true', dest='update',
                      help='Overwrite new and changed files', default=False)
    parser.add_option('-T', '--no-toc', action='store_true', dest='notoc',
                      help='Don\'t create a table of contents file')
    parser.add_option('-t', '--title', dest='toc_title', default='Javadoc',
                      help='Title to use on table of contents')
    parser.add_option('--no-member-headers', action='store_false', default=True, dest='member_headers',
                      help='Don\'t generate headers for class members')
    parser.add_option('-s', '--suffix', action='store', dest='suffix',
                      help='file suffix (default: rst)', default='rst')
    parser.add_option('-I', '--include', action='append', dest='includes',
                      help='Additional input paths to scan', default=[])
    parser.add_option('-p', '--parser', dest='parser_lib', default='lxml',
                      help='Beautiful Soup---html parser library option.')
    parser.add_option('-v', '--verbose', action='store_true', dest='verbose',
                      help='verbose output')

    (opts, args) = parser.parse_args(argv[1:])

    if not args:
        parser.error('A source path is required.')

    rootpath, excludes = args[0], args[1:]

    input_paths = opts.includes
    input_paths.append(rootpath)

    if not opts.destdir:
        parser.error('An output directory is required.')

    if opts.suffix.startswith('.'):
        opts.suffix = opts.suffix[1:]

    for input_path in input_paths:
        if not os.path.isdir(input_path):
            sys.stderr.write('%s is not a directory.\n' % (input_path,))
            sys.exit(1)

    if not os.path.isdir(opts.destdir):
        os.makedirs(opts.destdir)

    if opts.cache_dir and not os.path.isdir(opts.cache_dir):
        os.makedirs(opts.cache_dir)

    excludes = normalize_excludes(rootpath, excludes)
    source_files = []

    for input_path in input_paths:
        source_files.extend(find_source_files(input_path, excludes))

    packages, documents, sources = generate_documents(source_files, opts.cache_dir, opts.verbose,
                                                      opts.member_headers, opts.parser_lib)

    write_documents(packages, documents, sources, opts)

    if not opts.notoc:
        write_toc(packages, opts)



# Copyright 2021-2022. The MBI project. All rights reserved.
# This program is free software; you can redistribute it and/or modify it under the terms of the license (GNU GPL).

import os
import time
import subprocess
import sys
import re
import shlex
import select
import signal
import hashlib

class AbstractTool:
    def ensure_image(self, params="", dockerparams=""):
        """Verify that this is executed from the right docker image, and complain if not."""
        if os.path.exists("/MBI") or os.path.exists("trust_the_installation"):
            print("This seems to be a MBI docker image. Good.")
        else:
            print("Please run this script in a MBI docker image. Run these commands:")
            print("  docker build -f Dockerfile -t mpi-bugs-initiative:latest . # Only the first time")
            print(f"  docker run -it --rm --name MIB --volume $(pwd):/MBI {dockerparams}mpi-bugs-initiative /MBI/MBI.py {params}")
            sys.exit(1)

    def build(self, rootdir, cached=True):
        """Rebuilds the tool binaries. By default, we try to reuse the existing build."""
        print("Nothing to do to rebuild the tool binaries.")

    def setup(self, rootdir):
        """
        Ensure that this tool (previously built) is usable in this environment: setup the PATH, etc.
        This is called only once for all tests, from the logs directory.
        """
        # pass

    def run(self, execcmd, filename, binary, num_id, timeout, batchinfo):
        """Compile that test code and anaylse it with the Tool if needed (a cache system should be used)"""
        # pass

    def teardown(self):
        """
        Clean the results of all test runs: remove temp files and binaries.
        This is called only once for all tests, from the logs directory.
        """
        # pass

    def parse(self, cachefile):
        """Read the result of a previous run from the cache, and compute the test outcome"""
        return 'failure'

# Associate all possible detailed outcome to a given error scope. Scopes must be sorted alphabetically.
possible_details = {
    # scope limited to one call
    'InvalidBuffer':'AInvalidParam', 'InvalidCommunicator':'AInvalidParam', 'InvalidDatatype':'AInvalidParam', 'InvalidRoot':'AInvalidParam', 'InvalidTag':'AInvalidParam', 'InvalidWindow':'AInvalidParam', 'InvalidOperator':'AInvalidParam', 'InvalidOtherArg':'AInvalidParam', 'ActualDatatype':'AInvalidParam',
    'InvalidSrcDest':'AInvalidParam',
    # scope: Process-wide
#    'OutOfInitFini':'BInitFini',
    'CommunicatorLeak':'BResLeak', 'DatatypeLeak':'BResLeak', 'GroupLeak':'BResLeak', 'OperatorLeak':'BResLeak', 'TypeLeak':'BResLeak', 'RequestLeak':'BResLeak',
    'MissingStart':'BReqLifecycle', 'MissingWait':'BReqLifecycle',
    'LocalConcurrency':'BLocalConcurrency',
    # scope: communicator
    'CallMatching':'DMatch',
    'CommunicatorMatching':'CMatch', 'DatatypeMatching':'CMatch', 'OperatorMatching':'CMatch', 'RootMatching':'CMatch', 'TagMatching':'CMatch',
    'MessageRace':'DRace',

    'GlobalConcurrency':'DGlobalConcurrency',
    # larger scope
    'BufferingHazard':'EBufferingHazard',
    'OK':'FOK'}

error_scope = {
    'AInvalidParam':'single call',
    'BResLeak':'single process',
#    'BInitFini':'single process',
    'BReqLifecycle':'single process',
    'BLocalConcurrency':'single process',
    'CMatch':'multi-processes',
    'DRace':'multi-processes',
    'DMatch':'multi-processes',
    'DGlobalConcurrency':'multi-processes',
    'EBufferingHazard':'system',
    'FOK':'correct executions'
}

displayed_name = {
    'AInvalidParam':'Invalid parameter',
    'BResLeak':'Resource leak',
#    'BInitFini':'MPI call before initialization/after finalization',
    'BReqLifecycle':'Request lifecycle',
    'BLocalConcurrency':'Local concurrency',
    'CMatch':'Parameter matching',
    'DMatch':"Call ordering",
    'DRace':'Message race',
    'DGlobalConcurrency':'Global concurrency',
    'EBufferingHazard':'Buffering hazard',
    'FOK':"Correct execution",

    'aislinn':'Aislinn', 'civl':'CIVL', 'hermes':'Hermes', 'isp':'ISP', 'itac':'ITAC', 'simgrid':'Mc SimGrid', 'smpi':'SMPI', 'smpivg':'SMPI+VG', 'mpisv':'MPI-SV', 'must':'MUST', 'parcoach':'PARCOACH'
}

def parse_one_code(filename):
    """
    Reads the header of the provided filename, and extract a list of todo item, each of them being a (cmd, expect, test_num) tupple.
    The test_num is useful to build a log file containing both the binary and the test_num, when there is more than one test in the same binary.
    """
    res = []
    test_num = 0
    with open(filename, "r") as input_file:
        state = 0  # 0: before header; 1: in header; 2; after header
        line_num = 1
        for line in input_file:
            if re.match(".*BEGIN_MBI_TESTS.*", line):
                if state == 0:
                    state = 1
                else:
                    raise ValueError(f"MBI_TESTS header appears a second time at line {line_num}: \n{line}")
            elif re.match(".*END_MBI_TESTS.*", line):
                if state == 1:
                    state = 2
                else:
                    raise ValueError(f"Unexpected end of MBI_TESTS header at line {line_num}: \n{line}")
            if state == 1 and re.match(r'\s+\$ ?.*', line):
                m = re.match(r'\s+\$ ?(.*)', line)
                cmd = m.group(1)
                nextline = next(input_file)
                detail = 'OK'
                if re.match('[ |]*OK *', nextline):
                    expect = 'OK'
                else:
                    m = re.match('[ |]*ERROR: *(.*)', nextline)
                    if not m:
                        raise ValueError(
                            f"\n{filename}:{line_num}: MBI parse error: Test not followed by a proper 'ERROR' line:\n{line}{nextline}")
                    expect = 'ERROR'
                    detail = m.group(1)
                    if detail not in possible_details:
                        raise ValueError(
                            f"\n{filename}:{line_num}: MBI parse error: Detailled outcome {detail} is not one of the allowed ones.")
                test = {'filename': filename, 'id': test_num, 'cmd': cmd, 'expect': expect, 'detail': detail}
                res.append(test.copy())
                test_num += 1
                line_num += 1

    if state == 0:
        raise ValueError(f"MBI_TESTS header not found in file '{filename}'.")
    if state == 1:
        raise ValueError(f"MBI_TESTS header not properly ended in file '{filename}'.")

    if len(res) == 0:
        raise ValueError(f"No test found in {filename}. Please fix it.")
    return res

def categorize(tool, toolname, test_id, expected):
    outcome = tool.parse(test_id)

    if not os.path.exists(f'{test_id}.elapsed') and not os.path.exists(f'logs/{toolname}/{test_id}.elapsed'):
        if outcome == 'failure':
            elapsed = 0
        else:
            raise ValueError(f"Invalid test result: {test_id}.txt exists but not {test_id}.elapsed")
    else:
        with open(f'{test_id}.elapsed' if os.path.exists(f'{test_id}.elapsed') else f'logs/{toolname}/{test_id}.elapsed', 'r') as infile:
            elapsed = infile.read()

    # Properly categorize this run
    if outcome == 'timeout':
        res_category = 'timeout'
        if elapsed is None:
            diagnostic = 'hard timeout'
        else:
            diagnostic = f'timeout after {elapsed} sec'
    elif outcome == 'failure' or outcome == 'segfault':
        res_category = 'failure'
        diagnostic = 'tool error, or test not run'
    elif outcome == 'UNIMPLEMENTED':
        res_category = 'unimplemented'
        diagnostic = 'coverage issue'
    elif outcome == 'other':
        res_category = 'other'
        diagnostic = 'inconclusive run'
    elif expected == 'OK':
        if outcome == 'OK':
            res_category = 'TRUE_NEG'
            diagnostic = 'correctly reported no error'
        else:
            res_category = 'FALSE_POS'
            diagnostic = 'reported an error in a correct code'
    elif expected == 'ERROR':
        if outcome == 'OK':
            res_category = 'FALSE_NEG'
            diagnostic = 'failed to detect an error'
        else:
            res_category = 'TRUE_POS'
            diagnostic = 'correctly detected an error'
    else:
        raise ValueError(f"Unexpected expectation: {expected} (must be OK or ERROR)")

    return (res_category, elapsed, diagnostic, outcome)


def run_cmd(buildcmd, execcmd, cachefile, filename, binary, timeout, batchinfo, read_line_lambda=None):
    """
    Runs the test on need. Returns True if the test was ran, and False if it was cached.

    The result is cached if possible, and the test is rerun only if the `test.txt` (containing the tool output) or the `test.elapsed` (containing the timing info) do not exist, or if `test.md5sum` (containing the md5sum of the code to compile) does not match.

    Parameters:
     - buildcmd and execcmd are shell commands to run. buildcmd can be any shell line (incuding && groups), but execcmd must be a single binary to run.
     - cachefile is the name of the test
     - filename is the source file containing the code
     - binary the file name in which to compile the code
     - batchinfo: something like "1/1" to say that this run is the only batch (see -b parameter of MBI.py)
     - read_line_lambda: a lambda to which each line of the tool output is feed ASAP. It allows MUST to interrupt the execution when a deadlock is reported.
    """
    if os.path.exists(f'{cachefile}.txt') and os.path.exists(f'{cachefile}.elapsed') and os.path.exists(f'{cachefile}.md5sum'):
        hash_md5 = hashlib.md5()
        with open(filename, 'rb') as sourcefile:
            for chunk in iter(lambda: sourcefile.read(4096), b""):
                hash_md5.update(chunk)
        newdigest = hash_md5.hexdigest()
        with open(f'{cachefile}.md5sum', 'r') as md5file:
            olddigest = md5file.read()
        #print(f'Old digest: {olddigest}; New digest: {newdigest}')
        if olddigest == newdigest:
            print(f" (result cached -- digest: {olddigest})")
            return False
        os.remove(f'{cachefile}.txt')

    print(f"Wait up to {timeout} seconds")

    start_time = time.time()
    if buildcmd is None:
        output = f"No need to compile {binary}.c (batchinfo:{batchinfo})\n\n"
    else:
        output = f"Compiling {binary}.c (batchinfo:{batchinfo})\n\n"
        output += f"$ {buildcmd}\n"

        compil = subprocess.run(buildcmd, shell=True, check=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        if compil.stdout is not None:
            output += str(compil.stdout, errors='replace')
        if compil.returncode != 0:
            output += f"Compilation of {binary}.c raised an error (retcode: {compil.returncode})"
            for line in (output.split('\n')):
                print(f"| {line}", file=sys.stderr)
            with open(f'{cachefile}.elapsed', 'w') as outfile:
                outfile.write(str(time.time() - start_time))
            with open(f'{cachefile}.txt', 'w') as outfile:
                outfile.write(output)
            return True

    output += f"\n\nExecuting the command\n $ {execcmd}\n"
    for line in (output.split('\n')):
        print(f"| {line}", file=sys.stderr)

    # We run the subprocess and parse its output line by line, so that we can kill it as soon as it detects a timeout
    process = subprocess.Popen(shlex.split(execcmd), stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT, preexec_fn=os.setsid)
    poll_obj = select.poll()
    poll_obj.register(process.stdout, select.POLLIN)

    pid = process.pid
    pgid = os.getpgid(pid)  # We need that to forcefully kill subprocesses when leaving
    while True:
        if poll_obj.poll(5):  # Something to read? Do check the timeout status every 5 sec if not
            line = process.stdout.readline()
            # From byte array to string, replacing non-representable strings with question marks
            line = str(line, errors='replace')
            output = output + line
            print(f"| {line}", end='', file=sys.stderr)
            if read_line_lambda != None:
                read_line_lambda(line, process)
        if time.time() - start_time > timeout:
            with open(f'{cachefile}.timeout', 'w') as outfile:
                outfile.write(f'{time.time() - start_time} seconds')
            break
        if process.poll() is not None:  # The subprocess ended. Grab all existing output, and return
            line = 'more'
            while line != None and line != '':
                line = process.stdout.readline()
                if line is not None:
                    # From byte array to string, replacing non-representable strings with question marks
                    line = str(line, errors='replace')
                    output = output + line
                    print(f"| {line}", end='', file=sys.stderr)

            break

    # We want to clean all forked processes in all cases, no matter whether they are still running (timeout) or supposed to be off. The runners easily get clogged with zombies :(
    try:
        os.killpg(pgid, signal.SIGTERM)  # Terminate all forked processes, to make sure it's clean whatever the tool does
        process.terminate()  # No op if it's already stopped but useful on timeouts
        time.sleep(0.2)  # allow some time for the tool to finish its childs
        os.killpg(pgid, signal.SIGKILL)  # Finish 'em all, manually
        os.kill(pid, signal.SIGKILL)  # die! die! die!
    except ProcessLookupError:
        pass  # OK, it's gone now

    elapsed = time.time() - start_time

    rc = process.poll()
    if rc < 0:
        status = f"Command killed by signal {-rc}, elapsed time: {elapsed}\n"
    else:
        status = f"Command return code: {rc}, elapsed time: {elapsed}\n"
    print(status)
    output += status

    with open(f'{cachefile}.elapsed', 'w') as outfile:
        outfile.write(str(elapsed))

    with open(f'{cachefile}.txt', 'w') as outfile:
        outfile.write(output)
    with open(f'{cachefile}.md5sum', 'w') as outfile:
        hashed = hashlib.md5()
        with open(filename, 'rb') as sourcefile:
            for chunk in iter(lambda: sourcefile.read(4096), b""):
                hashed.update(chunk)
        outfile.write(hashed.hexdigest())

    return True



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: MBI

  Description: @{shortdesc}@
    @{longdesc}@

   Version of MPI: Conforms to MPI 1.1, does not require MPI 2 implementation

BEGIN_MPI_FEATURES
  P2P!basic: @{p2pfeature}@
  P2P!nonblocking: @{ip2pfeature}@
  P2P!persistent: Lacking
  COLL!basic: @{collfeature}@
  COLL!nonblocking: Lacking
  COLL!persistent: Lacking
  COLL!tools: Lacking
  RMA: Lacking
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 2 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>


int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;
  int dest, src;
  int root = 0;
  int stag = 0, rtag = 0;
  int buff_size = 1;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs < 2)
    printf("MBI ERROR: This test needs at least 2 processes to produce a bug!\\n");

  int dbs = sizeof(int)*nprocs; /* Size of the dynamic buffers for alltoall and friends */
  MPI_Comm newcom = MPI_COMM_WORLD;
  MPI_Datatype type = MPI_INT;
  MPI_Op op = MPI_SUM;

  @{init1}@
  @{init2}@
  @{init3}@
  if (rank == 0) {
    dest=1;src=1;
    @{operation3}@ /* MBIERROR1 */
    @{fini3}@
    @{operation1}@
    @{fini1}@
  }else if (rank==1) {
    dest=0;src=0;
    @{operation2}@ /* MBIERROR2 */
    @{fini2}@
    @{operation3}@
    @{fini3}@
  }

  @{free1}@
  @{free2}@
  @{free3}@

  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""


for s in gen.send + gen.isend:
    for r in gen.recv + gen.irecv:
        for c in gen.coll:
            patterns = {}
            patterns = {'s': s, 'r': r, 'c': c}
            patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
            patterns['p2pfeature'] = 'Yes' if s in gen.send or r in gen.recv else 'Lacking'
            patterns['ip2pfeature'] = 'Yes' if s in gen.isend or r in gen.irecv else 'Lacking'
            patterns['collfeature'] = 'Yes' if c in gen.coll else 'Lacking'
            patterns['s'] = s
            patterns['r'] = r
            patterns['c'] = c
            patterns['init1'] = gen.init[s]("1")
            patterns['init2'] = gen.init[r]("2")
            patterns['init3'] = gen.init[c]("3")
            patterns['fini1'] = gen.fini[s]("1")
            patterns['fini2'] = gen.fini[r]("2")
            patterns['fini3'] = gen.fini[c]("3")
            patterns['free1'] = gen.free[s]("1")
            patterns['free2'] = gen.free[r]("2")
            patterns['free3'] = gen.free[c]("3")
            patterns['operation1'] = gen.operation[s]("1")
            patterns['operation2'] = gen.operation[r]("2")
            patterns['operation3'] = gen.operation[c]("3")

            # Generate the incorrect matching because of the conditional
            replace = patterns.copy()
            replace['shortdesc'] = 'Point to point & collective mismatch'
            replace['longdesc'] = 'The @{s}@ corresponding to the @{r}@ is after @{c}@, while @{r}@ is before @{c}@. This is a deadlock.'
            replace['outcome'] = 'ERROR: CallMatching'
            replace['errormsg'] = 'P2P & Collective mistmatch. @{r}@ at @{filename}@:@{line:MBIERROR2}@ is matched with @{c}@ at @{filename}@:@{line:MBIERROR1}@ wich causes a deadlock.'
            gen.make_file(template, f'CallOrdering_{r}_{s}_{c}_nok.c', replace)

            # Generate the incorrect code depending on buffering
            #  replace = patterns.copy()
            #  replace['shortdesc'] = 'Point to point & collective mismatch'
            #  replace['longdesc'] = 'Point to point @{s}@ is matched with @{c}@ which causes a deadlock depending on the buffering mode.'
            #  replace['outcome'] = 'ERROR: BufferingHazard'
            #  replace['errormsg'] = 'P2P & Collective mistmatch. @{s}@ at @{filename}@:@{line:MBIERROR2}@ is matched with @{c}@ at @{filename}@:@{line:MBIERROR1}@ wich causes a deadlock.'
            #  replace['init1'] = init[s]("1")
            #  replace['init2'] = init[r]("2")
            #  replace['operation1'] = operation[r]("2")
            #  replace['operation2'] = operation[s]("1")
            #  replace['fini1'] = fini[r]("2")
            #  replace['fini2'] = fini[s]("1")
            #  gen.make_file(template, f'CollP2PBuffering_{r}_{s}_{c}_nok.c', replace)



#! /usr/bin/python3
import os
import sys
import generator_utils as gen

template = """// @{generatedby}@
/* ///////////////////////// The MPI Bugs Initiative ////////////////////////

  Origin: MBI

  Description: @{shortdesc}@
    @{longdesc}@

  Version of MPI: Conforms to MPI 1.1, does not require MPI 2 implementation

BEGIN_MPI_FEATURES
  P2P!basic: @{p2pfeature}@
  P2P!nonblocking: @{ip2pfeature}@
  P2P!persistent: Lacking
  COLL!basic: @{collfeature}@
  COLL!nonblocking: Lacking
  COLL!persistent: Lacking
  COLL!tools: Lacking
  RMA: Lacking
END_MPI_FEATURES

BEGIN_MBI_TESTS
  $ mpirun -np 4 ${EXE}
  | @{outcome}@
  | @{errormsg}@
END_MBI_TESTS
//////////////////////       End of MBI headers        /////////////////// */

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>


int main(int argc, char **argv) {
  int nprocs = -1;
  int rank = -1;
  int dest, src;
  int root = 0;
  int stag = 0, rtag = 0;
  int buff_size = 1;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  printf("Hello from rank %d \\n", rank);

  if (nprocs != 4)
    printf("MBI ERROR: This test needs 4 processes to produce a bug!\\n");

  int dbs = sizeof(int)*nprocs; /* Size of the dynamic buffers for alltoall and friends */
  MPI_Comm newcom = MPI_COMM_WORLD;
  MPI_Datatype type = MPI_INT;
  MPI_Op op = MPI_SUM;


  @{init1}@
  @{init2}@
  @{init3}@
  @{init4}@
  if (rank == 0) {
    dest=1;
    @{operation1}@
    @{fini1}@
    @{operation2}@
    @{fini2}@
  }else if (rank==2) {
    dest=1;
    @{operation1}@
    @{fini1}@
    @{operation2}@
    @{fini2}@
  }else if (rank==1) {
    src = MPI_ANY_SOURCE;
    rtag = MPI_ANY_TAG;
    @{operation3}@ /* MBIERROR1 */
    @{operation1}@
    @{fini1}@
    src = 0;
    @{operation4}@ /* MBIERROR2 */
    @{fini3}@
    @{fini4}@
  }else if (rank==3) {
    @{operation1}@
    @{fini1}@
  }

  @{free1}@
  @{free2}@
  @{free3}@
  @{free4}@

  MPI_Finalize();
  printf("Rank %d finished normally\\n", rank);
  return 0;
}
"""


for s in gen.send + gen.isend:
    for r in gen.irecv:
        for c in gen.coll:
            patterns = {}
            patterns = {'s': s, 'r': r, 'c': c}
            patterns['generatedby'] = f'DO NOT EDIT: this file was generated by {os.path.basename(sys.argv[0])}. DO NOT EDIT.'
            patterns['p2pfeature'] = 'Yes' if s in gen.send or r in gen.recv else 'Lacking'
            patterns['ip2pfeature'] = 'Yes' if s in gen.isend or r in gen.irecv else 'Lacking'
            patterns['collfeature'] = 'Yes' if c in gen.coll else 'Lacking'
            patterns['s'] = s
            patterns['r'] = r
            patterns['c'] = c
            patterns['init1'] = gen.init[c]("1")
            patterns['init2'] = gen.init[s]("2")
            patterns['init3'] = gen.init[r]("3")
            patterns['init4'] = gen.init[r]("4")
            patterns['fini1'] = gen.fini[c]("1")
            patterns['fini2'] = gen.fini[s]("2")
            patterns['fini3'] = gen.fini[r]("3")
            patterns['fini4'] = gen.fini[r]("4")
            patterns['free1'] = gen.free[c]("1")
            patterns['free2'] = gen.free[s]("2")
            patterns['free3'] = gen.free[r]("3")
            patterns['free4'] = gen.free[r]("4")
            patterns['operation1'] = gen.operation[c]("1")
            patterns['operation2'] = gen.operation[s]("2")
            patterns['operation3'] = gen.operation[r]("3")
            patterns['operation4'] = gen.operation[r]("4")

            # Generate the incorrect matching because of the conditional
            replace = patterns.copy()
            replace['shortdesc'] = 'Message race'
            replace['longdesc'] = 'Message race in @{r}@ with @{c}@.'
            replace['outcome'] = 'ERROR: MessageRace'
            replace['errormsg'] = 'Message race. The use of wildcard receive calls (@{r}@ at @{filename}@:@{line:MBIERROR1}@ and @{r}@ at @{filename}@:@{line:MBIERROR2}@) leads to nondeterministic matching.'
            gen.make_file(template, f'MessageRace_{c}_{s}_{r}_nok.c', replace)



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.

# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

"""
# ##################################################################################
# Take this tutorial online: https://simgrid.org/doc/latest/Tutorial_Algorithms.html
# ##################################################################################
"""

import sys
from simgrid import Engine, Mailbox, this_actor

# master-begin
def master(*args):
    if len(args) == 2:
        raise AssertionError(f"Actor master requires 4 parameters, but only {len(args)}")
    worker_count = int(args[0])
    tasks_count = int(args[1])
    compute_cost = int(args[2])
    communicate_cost = int(args[3])
    this_actor.info(f"Got {worker_count} workers and {tasks_count} tasks to process")

    for i in range(tasks_count): # For each task to be executed:
        # - Select a worker in a round-robin way
        mailbox = Mailbox.by_name(str(i % worker_count))

        # - Send the computation amount to the worker
        if (tasks_count < 10000 or (tasks_count < 100000 and i % 10000 == 0) or i % 100000 == 0):
            this_actor.info(f"Sending task {i} of {tasks_count} to mailbox '{mailbox.name}'")
        mailbox.put(compute_cost, communicate_cost)

    this_actor.info("All tasks have been dispatched. Request all workers to stop.")
    for i in range(worker_count):
        # The workers stop when receiving a negative compute_cost
        mailbox = Mailbox.by_name(str(i))
        mailbox.put(-1, 0)
# master-end

# worker-begin
def worker(*args):
    assert len(args) == 1, "The worker expects one argument"

    mailbox = Mailbox.by_name(args[0])
    done = False
    while not done:
        compute_cost = mailbox.get()
        if compute_cost > 0: # If compute_cost is valid, execute a computation of that cost
            this_actor.execute(compute_cost)
        else: # Stop when receiving an invalid compute_cost
            done = True

    this_actor.info("Exiting now.")
# worker-end

# main-begin
if __name__ == '__main__':
    assert len(sys.argv) > 2, f"Usage: python app-masterworkers.py platform_file deployment_file"

    e = Engine(sys.argv)

    # Register the classes representing the actors
    e.register_actor("master", master)
    e.register_actor("worker", worker)

    # Load the platform description and then deploy the application
    e.load_platform(sys.argv[1])
    e.load_deployment(sys.argv[2])

    # Run the simulation
    e.run()

    this_actor.info("Simulation is over")
# main-end



# Copyright (c) 2010-2022. The SimGrid Team. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the license (GNU LGPL) which comes with this package.

import sys

from simgrid import Engine, Actor, Comm, NetZone, Link, LinkInRoute, Mailbox, this_actor, NetworkFailureException


def sender(mailbox1_name: str, mailbox2_name: str) -> None:
    mailbox1: Mailbox = Mailbox.by_name(mailbox1_name)
    mailbox2: Mailbox = Mailbox.by_name(mailbox2_name)

    this_actor.info(f"Initiating asynchronous send to {mailbox1.name}")
    comm1: Comm = mailbox1.put_async(666, 5)

    this_actor.info(f"Initiating asynchronous send to {mailbox2.name}")
    comm2: Comm = mailbox2.put_async(666, 2)

    this_actor.info(f"Calling wait_any..")
    pending_comms = [comm1, comm2]
    try:
        index = Comm.wait_any([comm1, comm2])
        this_actor.info(f"Wait any returned index {index} (comm to {pending_comms[index].mailbox.name})")
    except NetworkFailureException:
        this_actor.info(f"Sender has experienced a network failure exception, so it knows that something went wrong")
        this_actor.info(f"Now it needs to figure out which of the two comms failed by looking at their state")

    this_actor.info(f"Comm to {comm1.mailbox.name} has state: {comm1.state_str}")
    this_actor.info(f"Comm to {comm2.mailbox.name} has state: {comm2.state_str}")

    try:
        comm1.wait()
    except NetworkFailureException:
        this_actor.info(f"Waiting on a FAILED comm raises an exception")

    this_actor.info("Wait for remaining comm, just to be nice")
    pending_comms.pop(0)
    try:
        Comm.wait_any(pending_comms)
    except Exception as e:
        this_actor.warning(str(e))


def receiver(mailbox_name: str) -> None:
    mailbox: Mailbox = Mailbox.by_name(mailbox_name)
    this_actor.info(f"Receiver posting a receive ({mailbox_name})...")
    try:
        mailbox.get()
        this_actor.info(f"Receiver has received successfully ({mailbox_name})!")
    except NetworkFailureException:
        this_actor.info(f"Receiver has experience a network failure exception ({mailbox_name})")


def link_killer(link_name: str) -> None:
    link_to_kill = Link.by_name(link_name)
    this_actor.info("sleeping 10 seconds...")
    this_actor.sleep_for(10.0)
    this_actor.info(f"turning off link {link_to_kill.name}")
    link_to_kill.turn_off()
    this_actor.info("link killed. exiting")


def main():
    e = Engine(sys.argv)
    zone: NetZone = NetZone.create_full_zone("AS0")
    host1 = zone.create_host("Host1", "1f")
    host2 = zone.create_host("Host2", "1f")
    host3 = zone.create_host("Host3", "1f")

    link_to_2 = LinkInRoute(zone.create_link("link_to_2", "1bps").seal())
    link_to_3 = LinkInRoute(zone.create_link("link_to_3", "1bps").seal())

    zone.add_route(host1.netpoint, host2.netpoint, None, None, [link_to_2], False)
    zone.add_route(host1.netpoint, host3.netpoint, None, None, [link_to_3], False)
    zone.seal()

    Actor.create("Sender", host1, sender, "mailbox2", "mailbox3")
    Actor.create("Receiver-1", host2, receiver, "mailbox2").daemonize()
    Actor.create("Receiver-2", host3, receiver, "mailbox3").daemonize()
    Actor.create("LinkKiller", host2, link_killer, "link_to_2").daemonize()

    e.run()


if __name__ == "__main__":
    main()
